from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.utils import icons
from kmviz.ui.components.factory import khide, kshow, km_color
from kmviz.ui.id_factory import kid
from kmviz.core.config import state
from kmviz.core.query import QueryResponseGeo
from dash_iconify import DashIconify
from dash_extensions.enrich import dcc, callback, Input, Output, State, html
from kmviz.ui.layouts.index import Index
from kmviz.ui.layouts.table import TableLayout
from kmviz.ui.layouts.sequence import SequenceLayout
from kmviz.ui.layouts.figure.plot import PlotLayout
from kmviz.ui.layouts.figure.map import MapLayout
from kmviz.ui.layouts.help import HelpLayout
from kmviz.ui.layouts.session import SessionLayout
from dash.exceptions import PreventUpdate
from flask import jsonify
import dash_mantine_components as dmc
import orjson
import dash_bootstrap_components as dbc
class Tabs:
    def __init__(self, st: state):
        self._tabs = []
        self._panels = []
        self.st = st

        self.dtab = None

        if self.st.mode == "session" or self.st.mode == "plot":
            self.dtab = "table"

        self.plugin_show = khide if self.st.mode in ("plot", "session") else {}

        self._index = Index(self.st)
        self._table = TableLayout(self.st)
        self._sequence = SequenceLayout(self.st)
        self._plot = PlotLayout(self.st, kid.plot, kid.plot["figure"], kid.table("grid"))
        self._map = MapLayout(self.st, kid.map, kid.map["figure"], kid.table("grid"))
        self._session = SessionLayout(self.st, kid.session)
        self._help = HelpLayout(self.st, kid.help)

        self._disabled = False if self.st.mode in ("plot", "session") else True

    def _plugin_layout(self):
        for name, plugin in self.st.conf.plugins.items():
            layouts = plugin.layouts()
            if layouts:
                for ptab, panel, icon in layouts:
                    if not isinstance(ptab, str):
                        tab_name = ptab[0]
                        tab = ptab[1]
                    else:
                        tab_name = ptab
                        tab = cf.tabs_tab(
                            kid.plugin[f"{name}-{tab_name}-tab"],
                            tab_name,
                            value=f"{name}-{tab_name}-tab",
                            leftSection=DashIconify(icon=icon) if icon else None,
                            style=self.plugin_show)

                    self._tabs.append(tab)
                    self._panels.append(
                        cf.tabs_panel(
                            kid.plugin[f"{name}-{tab_name}-panel"],
                            panel,
                            value=f"{name}-{tab_name}-tab",
                        )
                    )

    def _index_layout(self):

        self._tabs.append(
            cf.tabs_tab(
                kid.tabs["index"],
                "Index",
                value="index",
                leftSection=DashIconify(icon="iconoir:db"),
                disabled=self._disabled,
                style = khide.copy() if not self.st.ui.with_index_tab else {}
            )
        )
        self._panels.append(
            cf.tabs_panel(
                kid.tabs["index-panel"],
                self._index.layout(),
                value="index",
            )
        )

    def _index_callbacks(self) -> None:
        self._index.callbacks()

    def _table_layout(self):
        self._tabs.append(
            cf.tabs_tab(
                kid.tabs["table"],
                "Table",
                value="table",
                leftSection=DashIconify(icon="material-symbols:table"),
                disabled=self._disabled
            )
        )
        self._panels.append(
            cf.tabs_panel(
                kid.tabs["table-panel"],
                self._table.layout(),
                value="table"
            )
        )

    def _table_callbacks(self) -> None:
        self._table.callbacks()

    def _map_layout(self):
        self._tabs.append(
            cf.tabs_tab(
                kid.tabs["map"],
                "Map",
                value="map",
                leftSection=DashIconify(icon="fluent-mdl2:world"),
                disabled=self._disabled,
                style = khide.copy() if not self.st.ui.with_map_tab else {}
            )
        )
        self._panels.append(
            cf.tabs_panel(
                kid.tabs["map-panel"],
                self._map.layout(),
                value="map"
            )
        )

    def _map_callbacks(self) -> None:
        self._map.callbacks()

    def _plot_layout(self):
        self._tabs.append(
            cf.tabs_tab(
                kid.tabs["plot"],
                "Plot",
                value="plot",
                leftSection=DashIconify(icon="carbon:qq-plot"),
                disabled=self._disabled,
                style = khide.copy() if not self.st.ui.with_plot_tab else {}
            )
        )
        self._panels.append(
            cf.tabs_panel(
                kid.tabs["plot-panel"],
                self._plot.layout(),
                value="plot"
            )
        )

    def _plot_callbacks(self) -> None:
        self._plot.callbacks()

    def _sequence_layout(self):
        self._tabs.append(
            cf.tabs_tab(
                kid.tabs["sequence"],
                "Sequence",
                value="sequence",
                leftSection=DashIconify(icon="mdi:dna"),
                disabled=self._disabled,
                style = khide.copy() if not self.st.ui.with_sequence_tab else {}
            )
        )
        self._panels.append(
            cf.tabs_panel(
                kid.tabs["sequence-panel"],
                self._sequence.layout(),
                value="sequence"
            )
        )

    def _sequence_callbacks(self) -> None:
        self._sequence.callbacks()

    def _help_layout(self):
        self._tabs.append(
            cf.tabs_tab(
                kid.tabs["help"],
                "Help",
                value="help",
                leftSection=DashIconify(icon="material-symbols:help-outline"),
                style = khide.copy() if not self.st.ui.with_help_tab else {}
            )
        )
        self._panels.append(
            cf.tabs_panel(
                kid.tabs["help-panel"],
                self._help.layout(),
                value="help"
            )
        )

    def _help_callbacks(self) -> None:
        self._help.callbacks()

    def _corner_layout(self):
        style = kshow if self.st.mode != "plot" else khide

        group_content = [
            dcc.Download(id=kid.kmviz["download"]),
            dmc.Tooltip(
                label="When making figures, kmviz distinguishes 2 types of action: those that create a new figure (the 'Trace' tab for both maps and plots), e.g. change the plot type, and those that update the figure, e.g. change the title. The switch allows to enabled/disabled the auto updates. When enabled, all properties corresponding to update actions are automatically re-applied when the figure is modified. For example, if you set a title, then change the plot type, the title remains",
                withArrow=True,
                children=[cf.switch( kid.kmviz("auto"), onLabel=icons("autoff", width=20), offLabel=icons("auton", width=20), checked=True)],
                multiline=True,
                width=300
            ),
            dmc.Space(w=3, style=style),
            dmc.Tooltip(
                label="Download a session JSON file which can be loaded by any kmviz instance running in session mode. See kmviz documentation.",
                withArrow=True,
                children=[cf.action( kid.kmviz["download-button"], DashIconify(icon="bi:filetype-json", width=25), variant="filled", color = km_color, style=style)],
                multiline=True,
                width=300
            ),
            dmc.Space(w=5, style=style),
            cf.select( kid.kmviz("database"), data=[], placeholder="Database", size="xs", style=style),
            cf.select( kid.kmviz("query"), data=[], placeholder="Query", size="xs", style=style),
        ]
        self._tabs.append(cf.group(
            kid.kmviz["corner-grp"],
            *group_content,
            gap=4
        ))

    def _corner_callbacks(self) -> None:
        @callback(
            Input(kid.kmviz["download-button"], "n_clicks"),
            State(kid.store["results"], "data"),
            State(kid.store["session-id"], "data"),
            Output(kid.kmviz["download"], "data")
        )
        def download_all(n_clicks, data, session):
            if n_clicks and len(data) and self.st.engine.list():
                for qname, res in data.items():
                    for name in res:
                        R = QueryResponseGeo(
                            res[name]._query,
                            res[name]._response,
                            orjson.loads(res[name].df.to_json()),
                            _extra = {},
                            _geodata = self.st.engine.get(name).db.geodata
                        )
                        res[name] = R

                return dict(content=orjson.dumps(jsonify({session: data}).json).decode(), filename=f"session.json")
            raise PreventUpdate

    def _session_layout(self):
        self._tabs.append(
            self._session.layout()
        )

    def _session_callbacks(self):
        self._session.callbacks()

    def layout(self):
        if self.st.mode == "db":
            self._index_layout()

        self._table_layout()
        self._map_layout()
        self._plot_layout()

        if self.st.mode != "plot":
            self._sequence_layout()

        self._plugin_layout()
        self._help_layout()


        if self.st.mode == "session":
            self._session_layout()

        self._corner_layout()

        return cf.tabs(
            kid.tabs["tabs"],
            cf.tabs_list(
                kid.tabs["header"],
                *self._tabs
            ),
            *self._panels,
            value=self.dtab,
            allowTabDeactivation=True,
        )

    def callbacks(self):
        if self.st.ui.with_index_tab:
            self._index_callbacks()
        self._table_callbacks()

        if self.st.ui.with_map_tab:
            self._map_callbacks()
        if self.st.ui.with_plot_tab:
            self._plot_callbacks()
        if self.st.ui.with_sequence_tab:
            self._sequence_callbacks()

        self._corner_callbacks()

        if self.st.mode == "session":
            self._session_callbacks()



from kmviz.core import KmVizError
from kmviz.core.config import state
from kmviz.core.log import kmv_info

from kmviz.ui.components.factory import dmc_new
import dash

if dmc_new:
    dash._dash_renderer._set_react_version('18.2.0')

import dash_auth
from dash_extensions.enrich import DashProxy, LogTransform, NoOutputTransform, ServersideOutputTransform
from dash_extensions.enrich import html
import dash_mantine_components as dmc

from kmviz.api import KmvizAPI

import keyring


from dash_mantine_components import __version__ as dmc_version

def make_auth_function(usernames):
    def auth_function(username, password):
        if username not in usernames:
            return False
        p = keyring.get_password("kmviz", username)
        if p == password:
            return True
        return False
    return auth_function

def make_auth(application, config):
    if isinstance(config, dict):
        return dash_auth.BasicAuth(application, config)
    elif isinstance(config, list):
        return dash_auth.BasicAuth(application, auth_func=make_auth_function(set(config)))
    else:
        raise KmVizError("auth configuration is not valid")

def make_kmviz_app(st: state):
    app = DashProxy(
        __name__,
        transforms=[
            LogTransform(),
            NoOutputTransform(),
            ServersideOutputTransform(backends=[st.caches.server]),
        ],
        prevent_initial_callbacks=True,
        suppress_callback_exceptions=True,
        use_pages=True,
        pages_folder="ui/pages",
        background_callback_manager=st.caches.manager,
        external_stylesheets=st.css,
        external_scripts=st.js,
        meta_tags=st.html.metatags,
        compress=st.ui.compress_requests
    )

    if st.html.template:
        app.index_string = st.html.template

    if dmc_new:
        app.layout = dmc.MantineProvider([
            dash.page_container
        ])
    else:
        app.layout = html.Div([
            dash.page_container
        ])

    for page in dash.page_registry.values():
        kmv_info(f"📄 {page['name']} (path: '{page['path']}', relative_path: '{page['relative_path']}')")

    if st.auth:
        make_auth(app, st.auth)

    KmvizAPI(st, app.server)
    st.init_plugins_api(app.server)

    return app



from pathlib import Path

from dataclasses import dataclass
from pydantic import BaseModel, AfterValidator, ValidationInfo, ConfigDict, Field, PrivateAttr
from pydantic import model_validator, field_validator
from pydantic import PositiveInt, NonNegativeInt

from typing import List, Dict, Union, Annotated, Any, Optional, Literal, Tuple
from typing_extensions import Self

from kmviz.core import KmVizError
from kmviz.core.cache import KmvizResultCache
from kmviz.core.plugin import installed_plugins, search_for_plugin, copy_plugin_assets, copy_custom_assets
from kmviz.core.provider import Provider, Providers, PROVIDERS
from kmviz.core.metadata import MetaDB, METADBS
from kmviz.core.notifier import Notifier, NOTIFIERS, NullNotifier
from kmviz.core.query import Query, QueryResponse
from kmviz.ui.presets import cpresets

from functools import cached_property

class KmVizConfigError(KmVizError):
    pass

from kmviz.core.log import setup_logger
setup_logger()
from kmviz.core.log import kmv_info, kmv_ex, kmv_error, kmviz_idx

def wp_list(value: Any, empty: bool=False):
    if value is None:
        return list() if empty else None
    if not isinstance(value, list):
        return [value]
    return value

def parse_config(path: Union[str, dict]) -> dict:
    if isinstance(path, dict):
        return path
    else:
        path = Path(path)

    if not path.exists():
        raise KmVizConfigError(f"'{path}' does not exist!")

    res = {}
    stream = open(path, "rb")

    if path.suffix == ".json":
        import json
        res = json.load(stream)
    elif path.suffix == ".toml":
        import tomli
        res = tomli.load(stream)
    elif path.suffix == ".yaml" or path.suffix == ".yml":
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader
        from yaml import load as yload
        res = yload(stream, Loader=Loader)
    else:
        raise KmVizConfigError(f"'{path.name}' -> Supported format are 'json', 'toml' and 'yaml'")

    stream.close()
    return res

class chtml(BaseModel):
    template: str=""
    metatags: Annotated[
        Optional[
            Union[
                Dict[str, str],
                List[Dict[str, str]]
            ]
        ],
        AfterValidator(wp_list)
    ]=None

    @field_validator("template", mode="after")
    @classmethod
    def load_html_template(cls, value: str) -> str:
        if Path(value).exists():
            with open(value, "r") as fin:
                return fin.read()
        return ""

class cdefault(BaseModel):
    database: Annotated[
        Optional[
            Union[
                str,
                List[str]
            ]
        ],
        AfterValidator(wp_list)
    ]=None
    configuration: Optional[str]=None
    hide: Optional[bool]=False

    @model_validator(mode="after")
    def check(self) -> Self:
        if isinstance(self.database, str):
            self.database = [self.database]

        if self.database and not self.configuration:
            self.configuration = self.database[0]

        return self

class ccachelib_disk(BaseModel):
    cache_dir: str
    threshold: NonNegativeInt=0
    default_timeout: PositiveInt=86400
    model_config = ConfigDict(extra="allow")

class ccachelib_redis(BaseModel):
    host: str
    port: int=6379
    db: int=0
    default_timeout: PositiveInt=86400

    model_config = ConfigDict(extra="allow")

class cchachelib(BaseModel):
    type: Literal["disk", "redis"]
    params: Union[ccachelib_disk, ccachelib_redis]

    @model_validator(mode="after")
    def check(self) -> Self:
        if self.type == "disk" and not isinstance(self.params, ccachelib_disk):
            raise ValueError(f"{self.type}")
        if self.type == "redis" and not isinstance(self.params, ccachelib_redis):
            raise ValueError(f"{self.type}")
        return self

class ccachemanager_disk(BaseModel):
    directory: str
    model_config = ConfigDict(extra="allow")

class ccachemanager_celery(BaseModel):
    pass

class ccachemanager(BaseModel):
    type: Literal["disk", "celery"]
    params: Union[ccachemanager_disk, ccachemanager_celery]

    @model_validator(mode="after")
    def check(self) -> Self:
        if self.type == "celery":
            raise ValueError("celery manager is not implemented")
        return self

class ccache(BaseModel):
    result: Annotated[
        cchachelib,
        Field(
            default_factory=lambda: cchachelib(type="disk", params=ccachelib_disk(cache_dir=".kmviz_cache/result")),
            validate_default=True
        )
    ]

    server: Annotated[
        cchachelib,
        Field(
            default_factory=lambda: cchachelib(type="disk", params=ccachelib_disk(cache_dir=".kmviz_cache/server")),
            validate_default=True
        )
    ]

    manager: Annotated[
        ccachemanager,
        Field(
            default_factory=lambda: ccachemanager(type="disk", params=ccachemanager_disk(directory=".kmviz_cache/manager")),
            validate_default=True
        )
    ]

    @field_validator("result")
    @classmethod
    def make_result_cache(cls, value: cchachelib):
        from kmviz.core.cache import make_result_cache_manager, KmvizResultCache
        return KmvizResultCache(make_result_cache_manager(value.model_dump()))

    @field_validator("server")
    @classmethod
    def make_server_cache(cls, value: cchachelib):
        from kmviz.core.cache import make_server_cache_manager
        return make_server_cache_manager(value.model_dump())

    @field_validator("manager")
    @classmethod
    def make_manager_cache(cls, value: ccachemanager):
        from kmviz.core.cache import make_callback_cache_manager
        return make_callback_cache_manager(value.model_dump())

class cinput(BaseModel):
    max_query_size: PositiveInt=2**32
    max_size: PositiveInt=2**32
    max_query: PositiveInt=2**32
    alphabet: Literal["dna", "amino", "all"]="all"

class cmetadb(BaseModel):
    type: str
    params: Dict[str, Any]

class cnotif(BaseModel):
    type: str
    success: str
    failure: str
    subject: str
    subject_failure: str
    custom: Optional[Dict[str, Any]] = {}
    params: Optional[Dict[str, Any]] = {}

class cdatabase(BaseModel):
    type: str
    params: Dict[str, Any]
    defaults: Optional[Dict[str, Any]]=None
    metadata: cmetadb
    presets: Optional[str] = Field(None, validate_default=True)

    @field_validator("presets")
    @classmethod
    def make_presets(cls, value: Optional[str]) -> cpresets:
        if value:
            res = parse_config(value)
            return cpresets(**res)
        return cpresets(priority=False, map = {}, plot = {}, defaults=None)

PLOT_TYPES = [ "Scatter", "Line", "Area", "Bar", "Parallel categories", "Parallel coordinates", "Scatter matrix", "Density heatmap", "Density contour", "Violin", "Box", "Histogram" ]

class cui(BaseModel):
    with_map_tab: bool=True
    with_help_tab: bool=True
    with_plot_tab: bool=True
    with_sequence_tab: bool=True
    with_index_tab: bool=True
    crs: Optional[Dict[str, Tuple[str, Dict[str, Any]]]]={}
    ui_notif_msg: Optional[str] = ""
    compress_requests: Optional[bool]=True

    table_filter_message: Optional[str]="SQL WHERE clause"
    map_filter_message: Optional[str]=table_filter_message
    plot_filter_message: Optional[str]=table_filter_message

    plot_types: Optional[List[str]]=PLOT_TYPES

README_TEMPLATE = """
# {SESSION}

## {TABLE}

{FILES}

## Session

The file `session.json` allows to explore your results in any kmviz instance running in `session mode`.

```bash
pip install kmviz
python -m kmviz app start session
```

A kmviz instance is now running at `https://localhost:8050/dashboard`. You can upload your session file using the `Upload session` button.
"""

class capi(BaseModel):
    enabled: bool=False
    with_query: bool=True
    with_download: bool=True
    readme_template: Optional[str]=README_TEMPLATE
    route: Annotated[
        str,
        Field(default="/api", validate_default=True)
    ]
    query_route: Annotated[
        str,
        Field(default="/query", validate_default=True)
    ]
    download_route: Annotated[
        str,
        Field(default="/download", validate_default=True)
    ]
    limits: Annotated[
        cinput,
        Field(default=cinput())
    ]

    @field_validator("route", "query_route")
    @classmethod
    def check_route_syntax(cls, value: str, info: ValidationInfo):
        if value == "/":
            raise ValueError(f"'{info.field_name}' should not be '/'")
        if not value.startswith("/"):
            raise ValueError(f"'{info.field_name}' should start with '/'")
        if value.endswith("/"):
            raise ValueError(f"'{info.field_name}' ends with '/'")
        return value

def make_provider(name: str, config: cdatabase, providers: Dict[str, Provider]) -> Provider:
    if config.type not in providers:
        raise KmVizConfigError(f"Unknown provider type: '{config.type}'")

    prov = providers[config.type](name=name, **config.params)

    if config.defaults:
        prov.set_opt_defs(config.defaults)

    if config.presets:
        prov.set_presets(config.presets)

    return prov

def make_metadb(config: cmetadb, metadbs: Dict[str, MetaDB]) -> MetaDB:
    if config.type not in metadbs:
        raise KmVizConfigError(f"Unknown metadb type: '{config.type}'")
    return metadbs[config.type](**config.params)

def make_notif(config: cnotif, notifiers: Dict[str, Notifier]) -> Notifier:
    if config.type not in notifiers:
        raise KmVizConfigError(f"Unknown notifier type '{config.type}'")
    return notifiers[config.type](config.success, config.failure, config.subject, config.subject_failure, custom=config.custom, **config.params)

def make_db(name: str, config: cdatabase, providers: Dict[str, Provider], metadbs: Dict[str, MetaDB], connect: bool=True):
    try:
        prov = make_provider(name, config, providers)
    except Exception as e:
        raise KmVizConfigError(f"Error during '{name}' provider init: \"{e}\"")

    try:
        meta = make_metadb(config.metadata, metadbs)
    except Exception as e:
        raise KmVizConfigError(f"Error during '{name}' metadb init: \"{e}\"")

    prov.attach_metadata(meta)

    if connect:
        prov.db.connect()
        prov.connect()

    return prov


class ckmviz(BaseModel):
    idx: str = Field(kmviz_idx, frozen=True)
    databases: Dict[str, cdatabase]
    notif: Optional[cnotif]=None
    input: Optional[cinput]=cinput()
    default: cdefault = Field(cdefault(), alias="defaults")
    cache: Optional[ccache]=ccache()
    auth: Optional[Union[List[str], Dict[str, str]]]=None
    html: Optional[chtml]=chtml()
    api: Optional[capi]=capi()
    plugins: Optional[Dict[str, Dict[str, Any]]]={}
    assets: Optional[List[str]]=[]
    preset: Optional[Literal["flex", "fixed"]]="flex"
    ui: Optional[cui]=cui()

    _builtin_providers: Dict[str, Provider] = PrivateAttr(PROVIDERS.copy())
    _builtin_metadbs: Dict[str, Provider] = PrivateAttr(METADBS.copy())

    external_css: List[Union[Dict[str, Any], str]] = Field([], exclude=True)
    external_js: List[Union[Dict[str, Any], str]] = Field([], exclude=True)

    providers: Dict[str, Provider] = Field(PROVIDERS.copy(), exclude=True)
    metadbs: Dict[str, MetaDB] = Field(METADBS.copy(), exclude=True)
    notifiers: Dict[str, Notifier] = Field(NOTIFIERS.copy(), exclude=True)

    engine: Providers = Field(Providers(), exclude=True)
    notifier: Notifier = Field(NullNotifier(), exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("assets", mode="after")
    @classmethod
    def init_custom_assets(cls, value: Optional[List[str]]) -> Optional[List[str]]:
        copy_custom_assets(value)
        return value

    @field_validator("plugins", mode="after")
    @classmethod
    def init_plugins(cls, value: Optional[Dict[str, Dict[str, Any]]]) -> Optional[Dict[str, Dict[str, Any]]]:
        if value is None:
            return value

        installed = installed_plugins()
        res = {}
        for name, params in value.items():
            if not name.startswith("kmviz_"):
                raise ValueError(f"'{name}' is not a valid plugin name")
            if name not in installed:
                raise ValueError(f"'{name}' is not installed")

            kmv_info(f"💾 Init[{name}]: with {list(params.keys())}")
            plugin = search_for_plugin(name)
            plugin.configure(params)

            res[name] = plugin
            if path := copy_plugin_assets(name):
                kmv_info(f"💾 Assets[{name}]: from '{path}'")
        return res

    @model_validator(mode="after")
    def post_init_plugins(self) -> Self:
        for name, plugin in self.plugins.items():
            ps, ds, ns = [], [], []
            for pname, p in plugin.providers():
                if pname not in self.providers:
                    ps.append(pname)
                    self.providers[pname] = p
            for dname, d in plugin.databases():
                if dname not in self.metadbs:
                    ds.append(dname)
                    self.metadbs[dname] = d
            for nname, n in plugin.notifiers():
                if nname not in self.notifiers:
                    ns.append(nname)
                    self.notifiers[nname] = n

            kmv_info(f"💾 Provider[{name}]: include {ps}")
            kmv_info(f"💾 MetaDB[{name}]: include {ds}")
            kmv_info(f"💾 Notifier[{name}]: include {ns}")

            if styles := plugin.external_styles():
                self.external_css.extend(styles)
                kmv_info(f"💾 CSS[{name}]: include {styles}")
            if scripts := plugin.external_scripts():
                self.external_js.extend(scripts)
                kmv_info(f"💾 JS[{name}]: include {scripts}")

        return self

    @model_validator(mode="after")
    def init_databases(self) -> Self:
        for db_name in self.databases.keys():
            kmv_info(f"📖 Init[{db_name}]: with type '{self.databases[db_name].type}'")
            db = make_db(db_name, self.databases[db_name], self.providers, self.metadbs, True)
            self.engine.add(db)
        return self

    @model_validator(mode="after")
    def init_notifiers(self) -> Self:
        if self.notif:
            kmv_info(f"📖 Init[Notifier]: with type '{self.notif.type}'")
            self.notifier = make_notif(self.notif, self.notifiers)
        return self


@dataclass
class _wrap_cache:
    server: Any
    manager: Any

@dataclass
class _wrap_caches:
    cache: _wrap_cache

class state:
    def __init__(self, mode: Optional[str]="db", config: Optional[Union[dict, str]]=None):
        self._init = config
        self._mode: str=mode
        self._config: ckmviz=None

        if self._init:
            self.configure(self._init)

    def configure(self, mode: str, config: Union[dict, str], rethrow=False):
        try:
            self._init = config
            self._mode = mode
            if mode == "db":
                self._config = ckmviz(**parse_config(config))
            else:
                self._config = ckmviz(databases={})
        except Exception as e:
            if rethrow:
                raise e
            kmv_error(e)
            exit(1)

    @cached_property
    def instance_plugin(self) -> Tuple[Optional[str], Optional[Any]]:
        if not self._config.plugins:
            return "/dashboard", None

        for name, p in self._config.plugins.items():
            if p.is_instance_plugin():
                if "dashboard" in p.config and p.config["dashboard"] != "/":
                    return p.config["dashboard"], p
                else:
                    return "/dashboard", p
        return "/dashboard", None

    def init_plugins_api(self, app):
        if not self.api.enabled:
            return
        apis = []
        for name, p in self._config.plugins.items():
            if p.has_api():
                apis.append(name)
                p.set_api(app)
        if apis:
            kmv_info(f"Load APIs from: [{','.join(apis)}]")

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def auth(self) -> Optional[Union[List[str], Dict[str, str]]]:
        return self._config.auth

    @property
    def caches(self) -> ccache:
        return self._config.cache

    @property
    def cache(self) -> KmvizResultCache:
        return self._config.cache.result

    def put(self, idx: str, value: Any):
        self.cache.put(idx, value)

    def get(self, idx: str) -> Any:
        return self.cache.get(idx)

    @property
    def limits(self) -> cinput:
        return self._config.input

    @property
    def api(self) -> capi:
        return self._config.api

    @property
    def html(self) -> chtml:
        return self._config.html

    @property
    def defaults(self) -> cdefault:
        return self._config.default

    @property
    def ui(self) -> cui:
        return self._config.ui

    @property
    def databases(self) -> Providers:
        return self._config.engine

    @property
    def css(self) -> List[Union[Dict[str, Any], str]]:
        return self._config.external_css

    @property
    def js(self) ->  List[Union[Dict[str, Any], str]]:
        return self._config.external_js

    def database(self, name: str) -> Provider:
        return self._config.engine.get(name)

    @property
    def engine(self) -> Providers:
        return self._config.engine

    @property
    def notif(self) -> Notifier:
        return self._config.notifier

    @property
    def db_list(self) -> List[str]:
        return self._config.engine.list()

    def db_query(self, query: Query, actives: List[str], options: Dict[str, Any], idx: str) -> Dict[str, QueryResponse]:
        return self._config.engine.query(query, actives, options, idx)

    @property
    def conf(self) -> ckmviz:
        return self._config

def init_global_state():
    global st
    st = state()



from dash_extensions.enrich import callback, Input, State, Output
from kmviz.ui.id_factory import kid
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.core.config import state
from dash import html

class Global:
    def __init__(self, st: state):
        self.st = st

    def _store(self):

        default_options = {}
        for db_name in self.st.engine.list():
            default_options[db_name] = {}
            for opt_name, opt in self.st.engine.get(db_name).options.items():
                default_options[db_name][opt_name] = opt.value if opt.value else opt.default

        default_options["notif"] = {}
        for name, opt in self.st.notif.options().items():
            default_options["notif"][name] = opt.value if opt.value else opt.default

        return cf.div(
            kid.store["div"],
            cf.store(kid.pom["store"], data={}),
            cf.store(kid.store["session-id"], data=None),
            cf.store(kid.store("sequences"), data=[]),
            cf.store(kid.store("options"), data=default_options),
            cf.store(kid.store("databases"), data=self.st.defaults.database),
            cf.store(kid.store["state"], data={}),
            cf.store(kid.store["results"], data={}),
            cf.store(kid.store["presets"], data={}),
            html.Div(id=kid.submit["notification"]),
        )

    def _store_callbacks(self) -> None:
        @callback(
            Input(kid.store.all, "data"),
            State(kid.store.all, "id"),
            State(kid.store["state"], "data"),
            Output(kid.store["state"], "data"),
            Output(kid.submit["button"], "disabled"),
            prevent_initial_call=True
        )
        def is_ready(stores, ids, data):
            if "submit" in data and data["submit"]:
                return data, True

            r = {
                ids[index]["index"]: store is not None and len(store)
                for index, store in enumerate(stores)
            }

            if all(r.values()):
                r["submit"] = True
            else:
                r["submit"] = False

            return r, not r["submit"]

    def layout(self):
        return cf.div(
            kid.kmviz["global-div"],
            self._store()
        )

    def callbacks(self):
        self._store_callbacks()



from kmviz.core.config import state
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color, khide
from kmviz.ui.id_factory import kid
from kmviz.ui.utils import make_select_data
from kmviz.ui.components.helpers import load_file_content
from dash_extensions.enrich import Input, Output, callback, Serverside, html, no_update
from kmviz.core.query import Query, Response, QueryResponse
import dash_mantine_components as dmc


import pandas as pd
from dash_iconify import DashIconify
import base64
import orjson

class SessionLayout:
    def __init__(self, st: state, factory):
        self.st = state
        self.f = factory

    def layout(self):
        return cf.div(
            self.f["div"],
            cf.upload(
                self.f["session-file"],
                cf.button(
                    self.f["session-button"],
                    "Upload Session",
                    size="xs",
                    color=km_color,
                    leftSection=DashIconify(icon="bi:filetype-json", width=20),
                    className="kmviz-session-button"
                )
            ),
            cf.div(self.f["session-error"])
        )

    def callbacks(self):
        @callback(
            Input(self.f["session-file"], "filename"),
            Input(self.f["session-file"], "contents"),
            Output(kid.store["results"], "data"),
            Output(kid.pom["store"], "data"),
            Output(kid.kmviz("database"), "data"),
            Output(kid.kmviz("database"), "value"),
            Output(kid.kmviz("query"), "data"),
            Output(kid.kmviz("query"), "value"),
            Output(self.f["session-error"], "children"),
            Output(self.f["session-file"], "filename"),
            prevent_initial_call=True,
        )
        def input_session_file(filename, contents):
            try:
                content = load_file_content(filename, contents, "utf-8")
                session = orjson.loads(content)

                result = {}
                providers = []
                queries = []
                geo = {}

                session_id = list(session.keys())[0]
                session = session[session_id]
                for query, pres in session.items():
                    queries.append(query)
                    for provider, res in pres.items():
                        providers.append(provider)
                        q = Query(res["_query"]["_name"], res["_query"]["_seq"])
                        m = pd.DataFrame.from_dict(res["_metadata"])
                        m.insert(0, "ID", m.pop("ID"))
                        geo[provider] = res["_geodata"]
                        responses = {}
                        for sample, rep in res["_response"].items():
                            responses[sample] = Response(**rep)
                        result[query] = {provider: QueryResponse(q, responses, m)}
                pdata = make_select_data(providers)
                qdata = make_select_data(queries)

                return Serverside(result), {"session": geo}, pdata, providers[0], qdata, queries[0], [], None

            except Exception as e:
                msg = html.Div([dmc.Text("An error occured while loading your session file", color="red", weight=500)], className="kmviz-session-error")
                return (no_update,) * 6 + (msg, None)


from kmviz.core.config import state
from kmviz.core.io import parse_fastx, KmVizIOError
from kmviz.core.query import Query, QueryResponseGeo
from kmviz.core.log import kmv_info, kmv_warn
from flask import jsonify, send_file
from flask import request
import orjson
import uuid
import dataclasses
from io import StringIO, BytesIO
from zipfile import ZipFile

from typing import List

class KmvizAPI:
    def __init__(self, st: state, server):
        self.st = st
        self.server = server
        self._register()

    def _register(self):
        if self.st.api.enabled:
            if self.st.api.with_query:
                self._register_route(f"{self.st.api.route}{self.st.api.query_route}", ["POST"], self._make_query_callback())
                self._register_route(f"{self.st.api.route}{self.st.api.query_route}/<db>", ["POST"], self._make_query_metadata_callback())
            if self.st.api.with_download:
                self._register_route(f"{self.st.api.route}{self.st.api.download_route}/<session>", ["GET", "POST"], self._make_download_callback())


    def _get_options(self, database, form, with_prefix = True):
        options = {}
        for key in form:
            if with_prefix:
                if key.startswith(database):
                    n = key.split("#", 1)[1]
                    if n not in self.st.engine.get(database).options:
                        raise KmVizIOError(f"'{n}' is not a valid option for '{database}'")
                    options[n] = form[key]
            else:
                if key not in self.st.engine.get(database).options:
                    raise KmVizIOError(f"'{key}' is not a valid option for '{database}'")
                options[key] = form[key]

        for opt_name, opt in self.st.engine.get(database).options.items():
            if opt_name not in options:
                options[opt_name] = opt.value
            else:
                options[opt_name] = type(opt.value)(options[opt_name])
        return options

    def _make_geo_response(self, response, db, df_as_json):
        return QueryResponseGeo(
            response._query,
            response._response,
            orjson.loads(response.df.to_json()) if df_as_json else response.df,
            _extra = {},
            _geodata = self.st.engine.get(db).db.geodata
        )

    def _make_session_query(self, fastx: str, options: dict, df_as_json):
        uuid_str = f"kmviz-{str(uuid.uuid4())}"

        queries = [Query(name, seq) for name, seq, _ in parse_fastx(fastx, self.st.api.limits)]

        results = {}
        stores = {}

        for i, query in enumerate(queries):
            result = self.st.engine.query(query, list(options.keys()), options, uuid_str)

            stores[query.name] = {}
            for name in result:
                stores[query.name][name] = result[name]
                result[name] = self._make_geo_response(result[name], name, df_as_json)

            results[query.name] = result

        data_db = list(options.keys())
        def_db = data_db[0]

        self.st.put(uuid_str, (stores, data_db, def_db, [q.name for q in queries], queries[0].name))
        return { uuid_str: results }

    def _make_query_callback(self):
        def api_query():
            try:
                if not "database" in request.form:
                    kmv_warn("🔗 API (POST): 'database' is missing")
                    return f"'database' entry is missing!", 400
                if not "fastx" in request.files:
                    kmv_warn("🔗 API (POST): 'fastx' is missing")
                    return f"'fastx' entry is missing!", 400

                options = {}
                for db in request.form.getlist("database"):
                    options[db] = self._get_options(db, request.form, True)

                return jsonify(self._make_session_query(request.files["fastx"].stream.read().decode(), options, True))
            except KmVizIOError as e:
                kmv_warn(f"🔗 API (POST): Error ({str(e)})")
                return f"Error: {str(e)}", 400
            except Exception as e:
                kmv_warn(f"🔗 API (POST): Unknown error: {str(e)}")
                return "An error occured while processing your request", 400
        return api_query

    def _make_metadata_query(self, fastx: str, options: dict):
        uuid_str = f"kmviz-{str(uuid.uuid4())}"

        queries = [Query(name, seq) for name, seq, _ in parse_fastx(fastx, self.st.api.limits)]

        zf_io = BytesIO()
        with ZipFile(zf_io, "w") as zf:
            stores = {}
            for i, query in enumerate(queries):
                result = self.st.engine.query(query, list(options.keys()), options, uuid_str)
                stores[query.name] = {}
                for name in result:
                    zf.writestr(f"{query.name}.tsv", result[name].df.to_csv(index=False, sep="\t"))
                    stores[query.name][name] = result[name]


        data_db = list(options.keys())
        def_db = data_db[0]
        self.st.put(uuid_str, (stores, data_db, def_db, [q.name for q in queries], queries[0].name))

        zf_io.seek(0)
        return send_file(zf_io, download_name=f"{uuid_str}.zip")

    def _make_query_metadata_callback(self):
        def api_metadata_query(db):
            try:
                if db not in self.st.engine.list():
                    return f"'{db}' database does not exist!", 400
                if not "fastx" in request.files:
                    kmv_warn("🔗 API (POST): 'fastx' is missing")
                    return f"'fastx' entry is missing!", 400
                options = {db: self._get_options(db, request.form, False)}
                return self._make_metadata_query(request.files["fastx"].stream.read().decode(), options)
            except KmVizIOError as e:
                kmv_warn(f"🔗 API (POST): Error ({str(e)})")
                return f"Error: {str(e)}", 400
            except Exception as e:
                kmv_warn(f"🔗 API (POST): Unknown error: {str(e)}")
                return "An error occured while processing your request", 400
        return api_metadata_query


    def _make_info_callback(self):
        def api_info():
            results = dict(database={})

            kmv_info("🔗 API (GET)[info]")

            for name, db in self.st.engine.all().items():
                results["database"] = { name: {}}

                options = {}
                for opt_name, opt in db.options.items():
                    options[opt_name] = {
                        "type": type(opt.default).__name__,
                        "state": dataclasses.asdict(opt)
                    }
                results["database"][name]["options"] = options

            results["input"] = self.st.api.limits.model_dump()

            return jsonify(results)

        return api_info

    def _make_download_result(self, session):
        res = self.st.get(session)[0]
        zf_io = BytesIO()
        results = {}
        with ZipFile(zf_io, "w") as zf:
            for query_name, result in res.items():
                for name in result:
                    zf.writestr(f"{query_name}.tsv", result[name].df.to_csv(index=False, sep="\t"))
                    result[name] = self._make_geo_response(result[name], name, True)
                results[query_name] = result
            zf.writestr("session.json", orjson.dumps(jsonify({session: results}).json).decode())
            files = "\n".join(f"- `{k}.tsv`" for k in results)
            readme = self.st.api.readme_template.format(SESSION=session, FILES=files, TABLE="Tables" if len(results) > 1 else "Table")
            zf.writestr("README.md", readme)
        zf_io.seek(0)
        return send_file(zf_io, download_name=f"{session}.zip")

    def _make_download_callback(self):
        def api_download(session):
            try:
                kmv_info("🔗 API (POST)[download]")
                return self._make_download_result(session)
            except KmVizIOError as e:
                kmv_warn(f"🔗 API (GET): Error ({str(e)})")
                return f"Error: {str(e)}", 400
            except Exception as e:
                kmv_warn(f"🔗 API (GET): Unknown error: {str(e)}")
                return "An error occured while processing your request", 400
        return api_download

    def _register_route(self, route: str, methods: List[str], callback):
        self.server.route(route, methods=methods)(callback)



from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color
from kmviz.ui.utils import icons
from kmviz.ui.id_factory import IDFactory, kid
from dash_extensions.enrich import Input, Output, no_update, callback, ctx
from kmviz.ui.layouts.figure import patch_figure, patch_figure_all
import dash_mantine_components as dmc

class ColorLegendLayout:
    def __init__(self, factory: IDFactory, figure_id):
        self.f = factory
        self.fid = figure_id

    def _title_panel(self):
        return cf.tabs_panel(
            self.f["title-panel"],
            [
                cf.group(
                    self.f["title-grp-1"],
                    cf.text(self.f("colorbar_title_text"), label="Text", placeholder="<b>Title</b>", size="xs"),
                    cf.number(self.f("colorbar_title_font_size"), label="Font size", leftSection=icons("font_size"), min=0, max=100, value=None, step=1, placeholder="14", size="xs"),
                    cf.font(self.f("colorbar_title_font_family"), label="Font", size="xs"),
                    cf.select(self.f("colorbar_title_side"), data=["top", "bottom", "right"], label="Position", value="top", size="xs", leftSection=icons("pos"), className="kmviz-figure-select"),
                    cf.color(self.f("colorbar_title_font_color"), label="Font color", size="xs", value="#000000"),
                )
            ],
            value="title"
        )

    def _box_panel(self):
        orient = [
            {"label": "vertical", "value": "v"},
            {"label": "horizontal", "value": "h"},
        ]
        return cf.tabs_panel(
            self.f["box-panel"],
            [
                cf.group(
                    self.f["box-grp-1"],
                    cf.select(self.f("colorbar_xanchor"), label="Anchor X", data=["auto", "left", "center", "right"], value="auto", leftSection=icons("single"), size="xs"),
                    cf.select(self.f("colorbar_xref"), label="Reference X", data=["container", "paper"], value="container", leftSection=icons("single"), size="xs"),
                    cf.number(self.f("colorbar_x"), label="x", min=0.0, max=1.0, step=0.01, value=0.9, startValue=0.9, placeholder="0.5", decimalScale=2, leftSection=icons("floating"), size="xs"),
                    cf.select(self.f("colorbar_yanchor"), label="Anchor Y", data=["auto", "top", "middle", "bottom"], value="auto", leftSection=icons("single"), size="xs"),
                    cf.select(self.f("colorbar_yref"), label="Reference Y", data=["container", "paper"], value="container", leftSection=icons("single"), size="xs"),
                    cf.number(self.f("colorbar_y"), label="y", min=0.0, max=1.0, step=0.01, value=None, startValue=0.9, placeholder="0.95", decimalScale=2, leftSection=icons("floating"), size="xs"),
                    cf.number(self.f("colorbar_borderwidth"), label="Width", min=0, max=100, step=1, value=0, leftSection=icons("width", rotate=1), size="xs"),
                    cf.color(self.f("colorbar_bordercolor"), label="Border color", size="xs"),
                    cf.color(self.f("colorbar_bgcolor"), label="Bg color", size="xs"),
                )
            ],
            value="box"
        )

    def layout(self):
        return cf.tabs(
            self.f["legend"],
            cf.tabs_list(
                self.f["legend-tabslist"],
                cf.tabs_tab(self.f["title-tab"], "Title", value="title"),
                cf.tabs_tab(self.f["box-tab"], "Box", value="box"),
            ),
            self._title_panel(),
            self._box_panel(),
            value="title"
        )

    def callbacks(self):

        @callback(
            Output(self.fid, "figure", allow_duplicate=True),
            inputs=dict(inputs=(
                Input(self.f("colorbar_title_text"), "value"),
                Input(self.f("colorbar_title_font_size"), "value"),
                Input(self.f("colorbar_title_font_family"), "value"),
                Input(self.f("colorbar_title_font_color"), "value"),
                Input(self.f("colorbar_title_side"), "value"),
                Input(self.f("colorbar_borderwidth"), "value"),
                Input(self.f("colorbar_bordercolor"), "value"),
                Input(self.f("colorbar_bgcolor"), "value"),
                Input(self.f("colorbar_x"), "value"),
                Input(self.f("colorbar_xref"), "value"),
                Input(self.f("colorbar_xanchor"), "value"),
                Input(self.f("colorbar_y"), "value"),
                Input(self.f("colorbar_yref"), "value"),
                Input(self.f("colorbar_yanchor"), "value"),
            )),
            prevent_initial_call=True,
        )
        def update_legend(inputs):
            return patch_figure(ctx, ["layout", "coloraxis"])



import plotly.express as px
from kmviz.ui.utils import prevent_update_on_empty, prevent_update_on_none, PreventUpdate

px_bar_options = {
    "type",
    "xselect",
    "yselect",
    "color",
    "pattern_shape",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "hover_name",
    "hover_data",
    "custom_data",
    "text",
    "base",
    "error_x",
    "error_x_minus",
    "error_y",
    "error_y_minus",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "color_discrete_sequence",
    "color_discrete_map",
    "color_continuous_scale",
    "pattern_shape_sequence",
    "pattern_shape_map",
    "range_color",
    "color_continuous_midpoint",
    "opacity",
    "orientation",
    "barmode",
    "log_x",
    "log_y",
    "range_x",
    "range_y",
    "text_auto",
    "title",
    "template",
    "width",
    "height",
    "color_seq_continuous_scale",
    "color_div_continuous_scale",
    "color_cyc_continuous_scale",
}

px_histo_options = {
    "type",
    "xselect",
    "yselect",
    "color",
    "pattern_shape",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "hover_name",
    "hover_data",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "color_discrete_sequence",
    "color_discrete_map",
    "pattern_shape_sequence",
    "pattern_shape_map",
    "opacity",
    "orientation",
    "barmode",
    "log_x",
    "log_y",
    "range_x",
    "histfunc",
    "histnorm",
    "cumulative",
    "nbins",
    "range_y",
    "text_auto",
    "title",
    "template",
    "width",
    "height",
}

px_scatter_options = {
    "type",
    "xselect",
    "yselect",
    "size",
    "color",
    "text",
    "symbol",
    "animation_frame",
    "animation_group",
    "color_continuous_midpoint",
    "color_continuous_scale",
    "symbol_map",
    "symbol_sequence",
    "color_discrete_map",
    "color_discrete_sequence",
    "opacity",
    "size_max",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "template",
    "marginal_x",
    "marginal_y",
    "trendline",
    "trendline_scope",
    "trendline_options",
    "color_seq_continuous_scale",
    "color_div_continuous_scale",
    "color_cyc_continuous_scale",
}

px_line_options = {
    "type",
    "xselect",
    "yselect",
    "line_group",
    "color",
    "line_dash",
    "symbol",
    "hover_name",
    "hover_data",
    "custom_data",
    "text",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "error_x",
    "error_x_minus",
    "error_y",
    "error_y_minus",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "orientation",
    "color_discrete_sequence",
    "color_discrete_map",
    "line_dash_sequence",
    "line_dash_map",
    "symbol_sequence",
    "symbol_map",
    "markers",
    "log_x=",
    "log_y=",
    "range_x",
    "range_y",
    "line_shape",
    "render_mode",
    "title",
    "template",
    "width",
    "height"
}

px_area_options = {
    "type",
    "xselect",
    "yselect",
    "line_group",
    "color",
    "symbol",
    "hover_name",
    "hover_data",
    "custom_data",
    "text",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "error_x",
    "error_x_minus",
    "error_y",
    "error_y_minus",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "orientation",
    "color_discrete_sequence",
    "color_discrete_map",
    "symbol_sequence",
    "symbol_map",
    "markers",
    "log_x=",
    "log_y=",
    "range_x",
    "range_y",
    "line_shape",
    "render_mode",
    "title",
    "template",
    "width",
    "height"
}
px_scatter3d_options = {
    "type",
    "xselect",
    "yselect",
    "zselect",
}

px_pcoord_options = {
    "type",
    "dimensions",
    "color",
    "labels",
    "color_continuous_scale",
    "range_color",
    "color_continuous_midpoint",
    "title",
    "template",
    "width",
    "height",
    "color_seq_continuous_scale",
    "color_div_continuous_scale",
    "color_cyc_continuous_scale",

}

px_pcat_options = {
    "type",
    "dimensions",
    "color",
    "labels",
    "color_continuous_scale",
    "range_color",
    "color_continuous_midpoint",
    "title",
    "template",
    "width",
    "height",
    "dimensions_max_cardinality",
    "color_seq_continuous_scale",
    "color_div_continuous_scale",
    "color_cyc_continuous_scale",

}

px_pie_options = {
    "type",
    "names",
    "values",
    "color",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "color_discrete_sequence",
    "color_discrete_map",
    "hover_name",
    "hover_data",
    "custom_data",
    "category_orders",
    "labels",
    "title",
    "template",
    "width",
    "height",
    "opacity",
    "hole"
}

px_matrix_options = {
    "type",
    "dimensions",
    "color",
    "symbol",
    "size",
    "hover_name",
    "hover_data",
    "custom_data",
    "category_orders",
    "labels",
    "color_discrete_sequence",
    "color_discrete_map",
    "color_continuous_scale",
    "range_color",
    "color_continuous_midpoint",
    "symbol_sequence",
    "symbol_map",
    "opacity",
    "size_max",
    "title",
    "template",
    "width",
    "height",
    "color_seq_continuous_scale",
    "color_div_continuous_scale",
    "color_cyc_continuous_scale",

}

px_density_heat = {
    "xselect",
    "yselect",
    "zselect",
    "type",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "hover_name",
    "hover_data",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "orientation",
    "color_continuous_scale",
    "range_color",
    "color_continuous_midpoint",
    "marginal_x",
    "marginal_y",
    "log_x",
    "log_y",
    "range_x",
    "range_y",
    "histfunc",
    "histnorm",
    "nbinsx",
    "nbinsy",
    "text_auto",
    "title",
    "template",
    "width",
    "height",
    "color_seq_continuous_scale",
    "color_div_continuous_scale",
    "color_cyc_continuous_scale",

}

px_density_contour = {
    "xselect",
    "yselect",
    "zselect",
    "type",
    "color",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "hover_name",
    "hover_data",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "orientation",
    "color_discrete_sequence",
    "color_discrete_map",
    "marginal_x",
    "marginal_y",
    "trendline",
    "trendline_options",
    "trendline_color_override",
    "trendline_scope",
    "log_x",
    "log_y",
    "range_x",
    "range_y",
    "histfunc",
    "histnorm",
    "nbinsx",
    "nbinsy",
    "text_auto",
    "title",
    "template",
    "width",
    "contours_coloring",
    "height"
}

px_violin_options = {
    "xselect",
    "yselect",
    "type",
    "color",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "hover_name",
    "hover_data",
    "custom_data",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "color_discrete_sequence",
    "color_discrete_map",
    "orientation",
    "violinmode",
    "log_x",
    "log_y",
    "range_x",
    "range_y",
    "points",
    "box",
    "title",
    "template",
    "width",
    "height"
}

px_box_options = {
    "xselect",
    "yselect",
    "type",
    "color",
    "facet_row",
    "facet_col",
    "facet_col_wrap",
    "facet_row_spacing",
    "facet_col_spacing",
    "hover_name",
    "hover_data",
    "custom_data",
    "animation_frame",
    "animation_group",
    "category_orders",
    "labels",
    "color_discrete_sequence",
    "color_discrete_map",
    "orientation",
    "boxmode",
    "log_x",
    "log_y",
    "range_x",
    "range_y",
    "points",
    "notched",
    "title",
    "template",
    "width",
    "height"
}

px_options = {
    "Bar": px_bar_options,
    "Scatter": px_scatter_options,
    "Line": px_line_options,
    "Area": px_area_options,
    "Scatter3D": px_scatter3d_options,
    "Pie": px_pie_options,
    "Parallel coordinates": px_pcoord_options,
    "Parallel categories": px_pcat_options,
    "Scatter matrix": px_matrix_options,
    "Density heatmap": px_density_heat,
    "Density contour": px_density_contour,
    "Violin": px_violin_options,
    "Box": px_box_options,
    "Histogram": px_histo_options
}

def list_to_str(value):
    if isinstance(value, list) and len(value) == 1:
        return value[0]
    return value

def make_plot_px(ptype: str, df, X, Y, Z=None, params={}):
    X, Y, Z = list_to_str(X), list_to_str(Y), list_to_str(Z)

    if ptype == "Bar":
        return px.bar(df, x=X, y=Y, hover_name="ID", **params)

    if ptype == "Line":
        return px.line(df, x=X, y=Y, hover_name="ID", **params)

    if ptype == "Area":
        return px.area(df, x=X, y=Y, hover_name="ID", **params)

    if ptype == "Scatter":
        return px.scatter(df, x=X, y=Y, hover_name="ID", **params)

    if ptype == "Scatter3D":
        return px.scatter_3d(df, x=X, y=Y, z=Z, hover_name="ID", **params)

    if ptype == "Pie":
        return px.pie(df, hover_name="ID", **params),

    if ptype == "Parallel coordinates":
        prevent_update_on_empty(params["dimensions"])
        return px.parallel_coordinates(df, **params)

    if ptype == "Parallel categories":
        prevent_update_on_empty(params["dimensions"])
        return px.parallel_categories(df, **params)

    if ptype == "Scatter matrix":
        prevent_update_on_empty(params["dimensions"])
        return px.scatter_matrix(df, **params, hover_name="ID")

    if ptype == "Density contour":
        ct = None
        if "contours_coloring" in params:
            ct = params["contours_coloring"]
            del params["contours_coloring"]
        fig = px.density_contour(df, x=X, y=Y, z=Z, hover_name="ID", **params)

        if ct:
            fig.update_traces(contours_coloring=ct)

        return fig

    if ptype == "Density heatmap":
        return px.density_heatmap(df, x=X, y=Y, z=Z, hover_name="ID", **params)

    if ptype == "Violin":
        return px.violin(df, x=X, y=Y, hover_name="ID", **params)

    if ptype == "Box":
        return px.box(df, x=X, y=Y, hover_name="ID", **params)

    if ptype == "Histogram":
        return px.histogram(df, x=X, y=Y, hover_name="ID", **params)

def fix_px_params(params, ptype):
    for k in list(params):
        if k not in px_options[ptype]:
            del params[k]
    return params

def to_bool(value):
    return True if value == "True" else False

def select_cscale(scales):
    scale = None
    if any(scales):
        scale = next(s for s in scales if s is not None)
    return scale

def valid_input(ptype, X, Y, Z):

    if not ptype:
        raise PreventUpdate

    if ptype not in {"Parallel coordinates", "Parallel categories", "Pie", "Scatter matrix"}:
        prevent_update_on_empty(X, Y)

    if ptype in {"Density heatmap", "Density contour"}:
        prevent_update_on_empty(Z)





import os
import shutil
from pathlib import Path

from Bio import SeqIO

from kmviz.core import KmVizError
from io import StringIO

class KmVizIOError(KmVizError):
    pass

class KmVizInvalidQuery(KmVizIOError):
    pass

def is_fasta(content: str) -> bool:
    sio = StringIO(content)
    return any(SeqIO.parse(sio, "fasta"))

def is_fastq(content: str) -> bool:
    sio = StringIO(content)
    fq = SeqIO.parse(sio, "fastq")
    try:
        return any(fq)
    except Exception as e:
        return False

def input_type(content: str) -> str:
    if is_fasta(content):
        return "fasta"
    if is_fastq(content):
        return "fastq"

    raise KmVizIOError("Unsupported format, should be fasta/q.")

AMINO_ALPHABET = set("ABCDEFGHIKLMNPQRSTVWXYZabcdefghiklmnpqrstvwxyz")
DNA_ALPHABET = set("ACGTacgt")

def is_amino(content: str) -> bool:
    return all(c in AMINO_ALPHABET for c in content)

def is_dna(content: str) -> bool:
    return all(c in DNA_ALPHABET for c in content)

def validate_alphabet(seq, alpha):
    if alpha == "dna":
        return is_dna(seq)
    elif alpha == "amino":
        return is_amino(seq)
    else:
        return True

def validate_input(idx, seq, seq_size, max_size, nb, limits):
    if nb > limits.max_query:
        raise KmVizIOError(f"The max number of queries is {limits.max_query}")
    if seq_size > limits.max_query_size:
        raise KmVizIOError(f"Query max size is {limits.max_query_size} ({idx}:{seq_size})")
    if max_size > limits.max_size:
        raise KmVizIOError(f"Max cumulative query size is {limits.max_size} ({limits.max_size})")
    if not validate_alphabet(seq, limits.alphabet):
        raise KmVizIOError(f"Alphabet is not '{limits.alphabet}'")


def parse_fastx(content: str, limits: dict):
    sio = StringIO(content)
    max_size = 0
    nb = 0
    for record in SeqIO.parse(sio, input_type(content)):
        if len(record.seq) == 0:
            raise KmVizIOError("Empty sequence found.")

        seq_size = len(record.seq)
        max_size += seq_size
        nb += 1

        if limits is not None:
            validate_input(record.id, record.seq, seq_size, max_size, nb, limits)

        yield record.id, str(record.seq), None

def make_url(url: str, route: str=None, port: int=None) -> str:
    res = url
    if port:
        res += f":{port}"
    if route:
        res += route
    return res

def rm(*paths):
    for path in paths:
        p = Path(path)
        if p.is_file() or p.is_symlink():
            p.unlink()
        elif p.is_dir():
            shutil.rmtree(p)


from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color, dmc_old
from kmviz.ui.layouts.figure import patch_figure, patch_figure_all
from kmviz.ui.id_factory import IDFactory, kid
from kmviz.ui.utils import icons
from dash_iconify import DashIconify
import dash_mantine_components as dmc
from typing import Any
from dash_extensions.enrich import callback, no_update, Input, State, Output, ctx


class AxeLayout:
    def __init__(self, v: str, factory: IDFactory, figure_id):
        self.v = v
        self.fid = figure_id
        self.f = factory.new(self.v)

    def _select_axe(self, if_x: Any, if_y: Any):
        return if_x if self.v == "x" else if_y

    def _title_panel(self):
        return cf.tabs_panel(
            self.f["title-panel"],
            [
                cf.group(
                    self.f["title-grp-1"],
                    cf.text(self.f("title_text"), label="Text", placeholder="<b>Title</b>", size="xs", value=None),
                    cf.number(self.f("title_font_size"), label="Font size", leftSection=icons("font_size"), min=0, max=100, value=None, step=1, placeholder="14", size="xs"),
                    cf.font(self.f("title_font_family"), label="Font", size="xs"),
                    cf.color(self.f("title_font_color"), label="Font color", size="xs", value="#000000"),
                    cf.number(self.f("title_standoff"), label="Standoff", leftSection=icons("floating"), min=0, max=None, step=0.01, value=None, decimalScale=2, size="xs"),
                )
            ],
            value="title"
        )

    def _line_panel(self):
        return cf.tabs_panel(
            self.f["line-panel"],
            [
                cf.group(
                    self.f["line-grp-1"],
                    cf.number(self.f("linewidth"), label="Width", min=0, max=100, step=1, value=None, size="xs", leftSection=icons("width", rotate=1)),
                    cf.select(self.f("type"), label="Type", data=["linear", "log", "date", "category", "multicategory"], size="xs", leftSection=icons("axis"), className="kmviz-figure-select"),
                    cf.select(self.f("mirror"), label="Mirror", data=["True", "False", "ticks", "allticks"], size="xs", value="False", leftSection=icons("single"), className="kmviz-figure-select"),
                    cf.color(self.f("linecolor"), label="Line color", size="xs"),
                    cf.segmented(self.f("side"), data=self._select_axe(["top", "bottom"], ["left", "right"]), value=self._select_axe("bottom", "left"), size="xs", color=km_color, className="kmviz-figure-segmented"),
                    cf.switch(self.f("visible"), size="lg", onLabel="ON", offLabel="OFF", checked=True, className="kmviz-figure-switch"),
                )
            ],
            value="line"
        )

    def _zero_panel(self):
        return cf.tabs_panel(
            self.f["zero-panel"],
            [
                cf.group(
                    self.f["zero-grp-1"],
                    cf.number(self.f("zerolinewidth"), label="Width", min=0, max=None, step=1, value=1, leftSection=icons("width", rotate=1), size="xs"),
                    cf.color(self.f("zerolinecolor"), label="Line color", size="xs"),
                    cf.switch(self.f("zeroline"), size="lg", onLabel="ON", offLabel="OFF", checked=True, className="kmviz-figure-switch"),
                )
            ],
            value="zeroline"
        )

    def _grid_panel(self):
        return cf.tabs_panel(
            self.f["grid-panel"],
            [
                cf.group(
                    self.f["grid-grp-1"],
                    cf.number(self.f("gridwidth"), label="Width", min=0, max=None, step=1, value=None, leftSection=icons("width", rotate=1), size="xs"),
                    cf.select(self.f("griddash"), label="Style", data=["solid", "dot", "dash", "longdash", "dashdot", "longdashdot"], size="xs", value="solid", leftSection=icons("dash"), className="kmviz-figure-select"),
                    cf.color(self.f("gridcolor"), label="Grid color", size="xs"),
                    cf.switch(self.f("showgrid"), size="lg", onLabel="ON", offLabel="OFF", checked=True, className="kmviz-figure-switch"),
                )
            ],
            value="grid"
        )

    def _label_panel(self):
        return cf.tabs_panel(
            self.f["label-panel"],
            [
                cf.group(
                    self.f["label-grp-1"],

                    cf.number(self.f("tickangle"), label="Angle", min=-180, max=180, step=1, value=None, leftSection=icons("angle"), size="xs"),
                    cf.select(self.f("ticklabelposition"),
                              data=["outside", "inside", "outside top", "inside top", "outside left", "inside left", "outside right", "outside bottom", "inside bottom"],
                              size="xs", leftSection=icons("pos"), label="Position", className="kmviz-figure-select"),
                    cf.number(self.f("tickfont_size"), label="Font size", min=0, max=100, step=1, value=None, leftSection=icons("font_size"), size="xs"),
                    cf.font(self.f("tickfont_family"), label="Font", size="xs", className="kmviz-figure-select"),
                    cf.color(self.f("tickfont_color"), label="Label color", size="xs"),
                    cf.switch(self.f("showticklabels"), size="lg", onLabel="ON", offLabel="OFF", checked=True, className="kmviz-figure-switch"),
                    cf.switch(self.f("automargin"), size="lg", onLabel="Automargin", offLabel="Automargin", checked=True, className="kmviz-figure-switch"),

                )
            ],
            value="ticklabels"
        )

    def _marker_panel(self):
        return cf.tabs_panel(
            self.f["marker-panel"],
            [
                cf.group(
                    self.f["marker-grp-1"],

                    cf.number(self.f("ticklen"), label="Length", min=0, max=100, step=1, value=1, leftSection=icons("width", rotate=1), size="xs"),
                    cf.number(self.f("tickwidth"), label="Width", min=0, max=100, step=1, value=1, leftSection=icons("width"), size="xs"),
                    cf.number(self.f("nticks"), label="Max", min=0, max=10000, step=1, value=None, leftSection=icons("nindex"), size="xs"),
                    cf.color(self.f("tickcolor"), label="Tick color", size="xs"),
                    cf.segmented(self.f("ticks"), data=["inside", "outside", "hide"], value="hide", size="xs", color=km_color, className="kmviz-figure-segmented", fullWidth=True),
                )
            ],
            value="tickmarkers"
        )

    def layout(self):
        return cf.tabs(
            self.f["axe"],
            cf.tabs_list(
                self.f["axe-tabslist"],
                icons(self.v.upper(), color=km_color, width=30),
                cf.tabs_tab(self.f["title-tab"], "Title", value="title"),
                cf.tabs_tab(self.f["line-tab"], "Line", value="line"),
                cf.tabs_tab(self.f["zeroline-tab"], "Zeroline", value="zeroline"),
                cf.tabs_tab(self.f["grid-tab"], "Grid", value="grid"),
                cf.tabs_tab(self.f["tickslabels-tab"], "Tick Labels", value="ticklabels"),
                cf.tabs_tab(self.f["ticksmarkers-tab"], "Tick Markers", value="tickmarkers"),
                cf.number(self.f("axis-index"), min=0, max=20, step=1, value=0, leftSection=DashIconify(icon="tabler:number"), size="xs", style = {})#{"width": "75px", "height": "20px", "margin-top": "1px"}),
            ),
            self._title_panel(),
            self._line_panel(),
            self._zero_panel(),
            self._grid_panel(),
            self._label_panel(),
            self._marker_panel(),
            value=None,
            className="kmviz-fix-axes"
        )

    def callbacks(self):
        @callback(
            Output(self.fid, "figure"),
            inputs=dict(inputs=(
                State(self.f("axis-index"), "value"),
                Input(kid.kmviz("auto"), "checked"),
                Input(self.fid, "figure"),
                Input(self.f("title_text"), "value"),
                Input(self.f("title_font_size"), "value"),
                Input(self.f("title_font_color"), "value"),
                Input(self.f("title_font_family"), "value"),
                Input(self.f("title_standoff"), "value"),
                Input(self.f("visible"), "checked"),
                Input(self.f("type"), "value"),
                Input(self.f("side"), "value"),
                Input(self.f("linewidth"), "value"),
                Input(self.f("linecolor"), "value"),
                Input(self.f("mirror"), "value"),
                Input(self.f("showgrid"), "checked"),
                Input(self.f("griddash"), "value"),
                Input(self.f("gridwidth"), "value"),
                Input(self.f("gridcolor"), "value"),
                Input(self.f("showticklabels"), "checked"),
                Input(self.f("tickangle"), "value"),
                Input(self.f("ticklabelposition"), "value"),
                Input(self.f("tickfont_family"), "value"),
                Input(self.f("tickfont_size"), "value"),
                Input(self.f("tickfont_color"), "value"),
                Input(self.f("automargin"), "checked"),
                Input(self.f("ticks"), "value"),
                Input(self.f("ticklen"), "value"),
                Input(self.f("tickwidth"), "value"),
                Input(self.f("nticks"), "value"),
                Input(self.f("tickcolor"), "value"),
                Input(self.f("zeroline"), "checked"),
                Input(self.f("zerolinecolor"), "value"),
                Input(self.f("zerolinewidth"), "value"),
            )),
            prevent_initial_call=True,
        )
        def update_axis(inputs):
            if ctx.triggered_id == self.fid and not inputs[1]:
                return no_update

            skip = set([self.f("axis-index", True), self.fid, kid.kmviz("auto", True)])

            def make_patch(ax):
                n = f"{ax}axis" if inputs[0] == 0 else f"{ax}axis{inputs[0]+1}"
                if ctx.triggered_id == self.fid:
                    return patch_figure_all(ctx, ["layout", n], skip)
                else:
                    return patch_figure(ctx, ["layout", n])

            return make_patch(self.v)

class AxesLayout:
    def __init__(self, factory: IDFactory, figure_id):
        self.f = factory
        self.fid = figure_id
        self.x = AxeLayout("x", self.f, self.fid)
        self.y = AxeLayout("y", self.f, self.fid)

    def layout(self):
        if dmc_old:
            return cf.div(
                self.f["axes"],
                dmc.Grid([
                    dmc.Col(self.x.layout(), span=6),
                    dmc.Col(self.y.layout(), span=6),
                ])
            )
        else:
            return cf.div(
                self.f["axes"],
                dmc.Grid([
                    dmc.GridCol(self.x.layout(), span=6),
                    dmc.GridCol(self.y.layout(), span=6),
                ])
            )


    def callbacks(self):
        self.x.callbacks()
        self.y.callbacks()




from dash_extensions.enrich import Input, State, Output, callback, dcc
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color
from dash_iconify import DashIconify
import plotly.io as pio

class DownloadGraphLayout:
    def __init__(self, factory, figure_id, filename = "fig"):
        self.f = factory
        self.fid = figure_id
        self.filename = filename

    def layout(self):
        return cf.group(
            self.f["dl-grp"],
            dcc.Download(id=self.f["download"]),
            cf.segmented(
                self.f["format"],
                ["png", "jpg", "svg", "pdf", "html", "json"],
                value="png",
                size="xs",
                color=km_color
            ),
            cf.action(
                self.f["button"],
                DashIconify(icon="material-symbols:download", width=20),
                variant="filled",
                color=km_color,
            ),
            gap=3
        )

    def callbacks(self) -> None:
        @callback(
            Input(self.f["button"], "n_clicks"),
            State(self.f["format"], "value"),
            State(self.fid, "figure"),
            Output(self.f["download"], "data")
        )
        def download(n_clicks, fmt, data):
            if n_clicks:
                if fmt == "html":
                    content = pio.to_html(data, validate=False)
                    return dict(content=content, filename=f"{self.filename}.html")
                elif fmt == "json":
                    content = pio.to_json(data, validate=False)
                    return dict(content=content, filename=f"{self.filename}.json")
                else:
                    content = pio.to_image(data, fmt, validate=False)
                    return dcc.send_bytes(content, f"{self.filename}.{fmt}")



from kmviz.ui.id_factory import IDFactory, kid
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.layouts.figure import patch_figure, patch_figure_all
from dash_extensions.enrich import Input, State, Output, callback, ctx, no_update
from dash import html
from kmviz.ui.utils import icons

class TitleLayout:
    def __init__(self, factory: IDFactory, figure_id):
        self.f = factory
        self.fid = figure_id

    def layout(self) -> html.Div:
        return cf.div(
            self.f["div"],
            cf.group(
                self.f["grp-1"],
                cf.text(self.f("title_text"), label="Text", placeholder="<b>Title</b>", size="xs", value=None),
                cf.number(self.f("title_font_size"), label="Size", min=1, max=100, step=1, value=16, leftSection=icons("font_size"), size="xs"),
                cf.font(self.f("title_font_family"), label="Font", size="xs"),
                cf.color(self.f("title_font_color"), label="Font color", size="xs", value="#000000")
            ),
            cf.group(
                self.f["grp-2"],
                cf.select(self.f("title_xanchor"), label="Anchor X", data=["auto", "left", "center", "right"], value="auto", leftSection=icons("single"), size="xs"),
                cf.select(self.f("title_xref"), label="Reference X", data=["container", "paper"], value="container", leftSection=icons("single"), size="xs"),
                cf.number(self.f("title_x"), label="x", min=0.0, max=1.0, step=0.01, value=None, startValue=0.5, decimalScale=2, leftSection=icons("floating"), size="xs"),
                cf.select(self.f("title_yanchor"), label="Anchor Y", data=["auto", "top", "middle", "bottom"], value="auto", leftSection=icons("single"), size="xs"),
                cf.select(self.f("title_yref"), label="Reference Y", data=["container", "paper"], value="container", leftSection=icons("single"), size="xs"),
                cf.number(self.f("title_y"), label="y", min=0.0, max=1.0, step=0.01, value=None, startValue=0.9, decimalScale=2, leftSection=icons("floating"), size="xs")
            )
        )

    def callbacks(self) -> None:

        @callback(
            Output(self.fid, "figure", allow_duplicate=True),
            inputs=dict(inputs=(
                Input(self.fid, "figure"),
                Input(kid.kmviz("auto"), "checked"),
                Input(self.f("title_text"), "value"),
                Input(self.f("title_font_size"), "value"),
                Input(self.f("title_font_family"), "value"),
                Input(self.f("title_font_color"), "value"),
                Input(self.f("title_xanchor"), "value"),
                Input(self.f("title_xref"), "value"),
                Input(self.f("title_x"), "value"),
                Input(self.f("title_yanchor"), "value"),
                Input(self.f("title_yref"), "value"),
                Input(self.f("title_y"), "value"),
            )),
            prevent_initial_call=True

        )
        def update_title(inputs):
            skip = set([self.fid, kid.kmviz("auto", True)])
            if ctx.triggered_id == self.fid:
                if not inputs[1]:
                    return no_update
                return patch_figure_all(ctx, ["layout"], skip)
            return patch_figure(ctx)



import click
import sys
import os

from kmviz.core.log import kmv_info, kmv_error, kmv_warn, kmv_ex, kmv_debug
from kmviz.core.log import setup_logger
from kmviz.api import KmvizAPI

from kmviz.core import KmVizError
from typing import Any
import json

from kmviz.core.io import make_url
import kmviz.core.config as kconf
from kmviz.init import make_kmviz_app

from contextlib import redirect_stdout

from multiprocessing import cpu_count
from flask import Flask

from kmviz import __version__ as kmviz_version

import requests

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

def check_input_path(config, skip):
    if not skip and config is None:
        kmv_error("Path to config file is empty. Use 'kmviz app -c/--config <path>' or set KMVIZ_CONF env variable")
        exit(1)

def epilog(page: str=None) -> str:
    EPILOG = """
    \b
    ---
    kmviz v{v}
    Check out our documentation at {url}
    Contact: teo.lemane@genoscope.cns.fr
    ---
    """
    base="https://tlemane.github.io/tlemane/kmviz"
    return EPILOG.format(url=base if not page else f"{base}/page", v=kmviz_version)

@click.group(epilog=epilog())
@click.option("--verbose", "-v", type=click.Choice(["debug", "info", "warn", "error"]), help="Verbosity level", default="info", show_default=True)
@click.option("--log-dir", "-d", type=str, metavar="<dir>", default="", help="Path to log directory")
@click.option("--demo", "-x", type=str, default=None, hidden=True, help="Demo mode")
def kmviz(verbose, log_dir, demo):
    setup_logger(verbose, log_directory=log_dir, traceback=0 if verbose != "debug" else None)

    if demo is not None:
        import kmviz
        kmviz._set_version_str(demo)

@kmviz.command()
def version():
    """
    Show version and exit
    """
    import kmviz
    print(f"kmviz v{kmviz.__version__}")

@click.group(help="Query kmviz REST API", epilog=epilog())
@click.option("--url", "-u", default="http://localhost", help="Host", show_default=True, metavar="<str>")
@click.option("--port", "-p", default=None, help="Port", show_default=True, type=int, metavar="<int>")
@click.option("--auth", "-a", default=None, help="HTTP BasicAuth credentials", metavar="<user>:<password>")
@click.pass_context
def api(ctx, **kwargs):
    ctx.ensure_object(AttrDict)
    ctx.obj = AttrDict(**kwargs)

    if ctx.obj.auth:
        try:
            ctx.obj.auth = tuple(ctx.obj.auth.split(":", 1))
        except:
            kmv_error("Invalid auth format")
            exit(1)

    auth_text = ctx.obj.auth[0] if ctx.obj.auth else None
    kmv_info("url={url}, port={port}, auth={auth_text}", auth_text=auth_text, **kwargs)

@api.command(epilog=epilog())
@click.option("--route", "-r", default="/api", help="Route", show_default=True, metavar="<str>")
@click.option("--output", "-o", default="stdout", help="Output file", show_default=True, metavar="<path>")
@click.pass_context
def infos(ctx, route, output):
    """
    Get databases information
    """
    url = make_url(url=ctx.obj.url, port=ctx.obj.port, route=route)

    try:
        res = requests.get(url=url, allow_redirects=True, auth=ctx.obj.auth, headers={'content-type': 'application/json'})
    except Exception as e:
        kmv_error(e)
        exit(1)

    if res.status_code != 200:
        kmv_error(f"GET [{res.status_code}]: {res.content.decode()}")
        exit(1)

    kmv_info(f"GET {url} -> status {res.status_code}")

    if output != "stdout":
        with open(output, "w") as f:
            with redirect_stdout(f):
                print(json.dumps(res.json(), indent=4))
    else:
        print(json.dumps(res.json(), indent=4))

@api.command(epilog=epilog())
@click.option("--route", "-r", default="/api/query", help="Query route", show_default=True, metavar="<str>")
@click.option("--output", "-o", default="stdout", help="Output file", show_default=True, metavar="<path>")
@click.option("--database", "-d", required=True, multiple=True, help="Database to query", metavar="<str>")
@click.option("--opt", "-x", nargs=3, multiple=True, help="Database options", metavar="<db_name> <opt_name> <value>")
@click.option("--fastx", "-f", required=True, help="Path to fastx query file", metavar="<path>")
@click.pass_context
def query(ctx, route, output, database, opt, fastx):
    """
    Make queries and get results as a json session file, viewable using a local kmviz instance. See kmviz app start session
    """
    data = []
    for db in database:
        data.append(("database", db))

    for db, o, v in opt:
        if db not in database:
            raise KmVizError(f"Set '{o}' for '{db}', but '{db}' not in {database}")
        data.append((f"{db}#{o}", v))

    url = make_url(url=ctx.obj.url, port=ctx.obj.port, route=route)
    try:
        res = requests.post(url=url, data=data, files={"fastx": open(fastx, "rb")}, allow_redirects=True, auth=ctx.obj.auth)
    except Exception as e:
        kmv_error(e)
        exit(1)

    if res.status_code != 200:
        kmv_error(f"GET [{res.status_code}]: {res.content.decode()}")
        exit(1)

    kmv_info(f"POST {url} -> status {res.status_code}")

    if output != "stdout":
        with open(output, "w") as f:
            with redirect_stdout(f):
                print(json.dumps(res.json(), indent=4))
        kmv_info(output)
    else:
        print(json.dumps(res.json(), indent=4))


@api.command(epilog=epilog())
@click.option("--route", "-r", default="/api/query", help="Query route", show_default=True, metavar="<str>")
@click.option("--output", "-o", default=os.getcwd(), help="Output directory", show_default=True, metavar="<dir>")
@click.option("--database", "-d", required=True, help="Database to query", metavar="<str>")
@click.option("--opt", "-x", nargs=2, multiple=True, help="Database option, -x <opt_name> <value>", metavar="<opt_name> <value>")
@click.option("--fastx", "-f", required=True, help="Path to fastx query file", metavar="<path>")
@click.pass_context
def metadata(ctx, route, output, database, opt, fastx):
    """
    Make queries and get results as a tsv dataframe
    """
    data = []

    db_route = f"{route}/{database}"

    for o, v in opt:
        data.append((o, v))

    url = make_url(url=ctx.obj.url, port=ctx.obj.port, route=db_route)

    try:
        res = requests.post(url=url, data=data, files={"fastx": open(fastx, "rb")}, allow_redirects=True, auth=ctx.obj.auth)
    except Exception as e:
        kmv_error(e)
        exit(1)

    if res.status_code != 200:
        kmv_error(f"POST [{res.status_code}]: {res.content.decode()}")
        exit(1)

    kmv_info(f"POST {url} -> status {res.status_code}")

    filename = res.headers["Content-Disposition"].split("filename=")[1]

    with open(filename, "wb") as fout:
        fout.write(res.content)

    kmv_info(filename)

@click.group(epilog=epilog())
@click.option("--config", "-c",
              help="Path to config file",
              show_default=bool(os.environ.get("KMVIZ_CONF")),
              envvar="KMVIZ_CONF",
              show_envvar=True,
              metavar="<path>")
@click.pass_context
def app(ctx, **kwargs):
    """
    Start kmviz
    """
    ctx.ensure_object(AttrDict)
    ctx.obj = AttrDict(**kwargs)

@app.command(epilog=epilog())
@click.argument("mode", type=click.Choice(["db", "session", "plot", "api"]), default="db")
@click.option("--url", "-u", default="localhost", help="url", show_default=True, metavar="<str>")
@click.option("--port", "-p", default=8050, help="port", show_default=True, type=int, metavar="<int>")
@click.option("--no-seq-tab", is_flag=True)
@click.option("--debug", "-d", is_flag=True)
@click.pass_context
def start(ctx, mode, url, port, no_seq_tab, debug):
    """
    Start with flask server (recommended for single users)

    \b
    Mode:
      - db (default)
      - session: Session mode, see https://tlemane/github.io/kmviz/)
      - plot: Plot mode, see https://tlemane/github.io/kmviz/)
      - api: Only start the kmviz REST API (see https://tlemane/github.io/kmviz/)
    """
    kconf.init_global_state()
    check_input_path(ctx.obj.config, mode == "session" or mode == "plot")
    kconf.st.configure(mode if mode != "api" else "db", ctx.obj.config)

    if no_seq_tab:
        kconf.st.ui.with_sequence_tab = False

    if mode == "api":
        app = Flask(__name__)
        kconf.st._config.api.enabled = True
        KmvizAPI(kconf.st, app)
    else:
        app = make_kmviz_app(kconf.st)

    app.run(host=url, port=port, debug=debug)


import gunicorn.app.base

class GunicornApp(gunicorn.app.base.BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or None
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application

@app.command(epilog=epilog())
@click.option("--mode", "-m", default="db", type=click.Choice(["db", "api"]), show_default=True, metavar="<str>", help="Mode, 'db' or 'api'")
@click.option("--url", "-u", default="localhost", help="Host", show_default=True, metavar="<str>")
@click.option("--port", "-p", default=8000, help="Port", show_default=True, type=int, metavar="<int>")
@click.option("--workers", "-w", default=cpu_count() * 2 + 1, help="Number of workers", show_default=True, type=int, metavar="<int>")
@click.option("--reload-extra-file", "-r", multiple=True, type=str, metavar="<str>", help="Reload when files change")
@click.option("--extra", "-e", multiple=True, type=(str, str), help="Extra gunicorn options",metavar="<option> <value>")
@click.pass_context
def deploy(ctx, mode, url, port, workers, reload_extra_file, extra):
    """
    Deployment with gunicorn (recommended for multi-user instances)
    """
    options = {
        "bind": f"{url}:{port}",
        "workers": workers,
        "reload_extra_files": reload_extra_file,
        "reload": True if reload_extra_file else False
    }

    for opt in extra:
        if opt[0] not in options:
            options[opt[0]] = opt[1]

    kconf.init_global_state()
    kconf.st.configure("db", ctx.obj.config)

    if mode == "api":
        app = Flask(__name__)
        kconf.st._config.api.enabled = True
        KmvizAPI(kconf.st, app)
    else:
        app = make_kmviz_app(kconf.st)
    GunicornApp(app.server, options).run()


@click.group(epilog=epilog())
def config():
    """
    Configuration helpers
    """
    pass

@config.command(epilog=epilog())
@click.option("--config", "-c", required=os.environ.get("KMVIZ_CONF", None) is None,
              help="Path to config file",
              default=lambda: os.environ.get("KMVIZ_CONF"),
              show_default=os.environ.get("KMVIZ_CONF"),
              metavar="<path>")
@click.option("--with-init", "-w", is_flag=True, help="Try to init")
@click.option("--details", "-d", is_flag=True, help="Show schema on error")
def check(config, with_init, details):
    """
    Valid configuration file.
    \b
    With --with-init, the validation process also tries to init the app,
    i.e. connect to databases, load plugins, etc
    """
    from kmviz.core.config import ckmviz, parse_config
    from jsonschema import validate
    schema = ckmviz.model_json_schema(schema_generator=CustomJsonSchema)
    instance = parse_config(config)

    try:
        validate(instance=instance, schema=schema)
    except Exception as e:
        if details:
            kmv_error(e)
        else:
            kmv_error(e.message)
        exit(1)

    kmv_info("✅ Schema ok!")

    if with_init:
        try:
            setup_logger("ERROR")
            kconf.init_global_state()
            kconf.st.configure("db", config, True)
            setup_logger("DEBUG")
            kmv_info("✅ Init ok!")
        except Exception as e:
            kmv_error(f"❌ Init fails: '{str(e)}'")
            exit(1)


from pydantic_core import PydanticOmit, core_schema
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue

class CustomJsonSchema(GenerateJsonSchema):
    def handle_invalid_for_json_schema(
        self, schema: core_schema.CoreSchema, error_info: str
    ) -> JsonSchemaValue:
        raise PydanticOmit

@config.command(epilog=epilog())
@click.option("--output", "-o", default="stdout", help="Output file prefix", show_default=True, metavar="<path>")
def schema(fmt, output):
    """
    Make JSON schema
    """
    from kmviz.core.config import ckmviz
    schema = json.dumps(ckmviz.model_json_schema(schema_generator=CustomJsonSchema), indent=4)

    if output != "stdout":
        with open(f"{output}.json", "r") as f:
            with redirect_stdout(f):
                print(schema)
    else:
        print(schema)

def convert_to(data: dict, format: str):
    import tomli_w
    from yaml import Dumper, dump

    if format == "json":
        return json.dumps(data, indent=4)
    elif format == "toml":
        return tomli_w.dumps(data)
    elif format == "yaml":
        return dump(data, Dumper=Dumper)

from kmviz.core.config import *

@config.command(epilog=epilog())
@click.option("--fmt", "-f", type=click.Choice(["toml", "yaml", "json"]), default="yaml", help="File format", show_default=True)
@click.option("--section", "-s", type=click.Choice(["all", "databases", "input", "api", "auth", "default", "plugins", "cache", "html"]), multiple=True, default=["all"], show_default=True)
@click.option("--output", "-o", default="stdout", help="Output file prefix", show_default=True, metavar="<path>")
def template(fmt, section, output):
    """
    Make configuration templates
    """
    if "all" in section:
        section=None

    cache = ccache.model_construct(
        result=cchachelib(type="disk", params=ccachelib_disk(cache_dir=".kmviz/result")),
        server=cchachelib(type="disk", params=ccachelib_disk(cache_dir=".kmviz/server")),
        manager=ccachemanager(type="disk", params=ccachemanager_disk(directory=".kmviz/manager")),
    )

    html = chtml.model_construct(
        template="template.html",
        metatags=dict(description="kmviz ex description")
    )

    databases = {
        "ExampleDB": cdatabase.model_construct(
            type="kmindex-server",
            params={"url": "127.0.0.1", "port": 8080},
            metadata=cmetadb.model_construct(
                type="tsv",
                params={
                    "path": "./metadata.tsv",
                    "sep": "\t",
                    "idx": "Sample",
                    "geodata": {"latitude": "Lat", "longitude":"Long"}
                }
            ),
            defaults={
                "z" : {"value": 2, "hide": True},
                "coverage" : {"value": 0.1, "hide": False}
            },
            presets="./presets.toml"
        )
    }

    default = cdefault(database="ExDB", configuration="ExampleDB")
    plugins = { "kmviz_example": {}}

    auth = {"ex_user": "ex_password"}

    c = ckmviz.model_construct(
        databases=databases,
        default=default,
        cache=cache,
        html=html,
        plugins=plugins,
        auth=auth
    ).model_dump(include=section, exclude="idx")

    if output != "stdout":
        with open(f"{output}.{fmt}", "w") as f:
            with redirect_stdout(f):
                print(schema)
    else:
        print(convert_to(c, fmt))


@click.group(epilog=epilog())
def plugin():
    """
    Plugin helpers
    """
    pass

@plugin.command(epilog=epilog())
def list():
    """
    List installed plugins
    """
    from kmviz.core.plugin import installed_plugins
    print(installed_plugins())

groups = [
    "kmviz",
    "store",
    "side",
    "input",
    "index",
    "table",
    "map",
    "plot",
    "sequence",
    "config"
]

@plugin.command("list-id", hidden=True)
@click.option("--group", "-g", type=click.Choice(groups), required=True, help="Identifier group")
@click.option("--fmt", "-f", is_flag=True, help="Formatted output")
@click.option("--string", "-s", is_flag=True, help="With string ids")
@click.option("--index", "-i", is_flag=True, help="With dict ids")
def idx(group, fmt, string, index):
    """
    List kmviz component identifiers

    \b
    This command is hidden in kmviz help and mainly
    used as a helper for writing documentation
    """
    from kmviz.ui.id_factory import kid, _split_idx, _formatted_idx
    from kmviz.ui.layouts import Global
    from kmviz.ui.layouts.sidebar import Sidebar
    from kmviz.ui.layouts.tabs import Tabs

    kconf.init_global_state()
    kconf.st.configure(None, None)
    kconf.st._mode = "db"

    Global(kconf.st).layout()
    Sidebar(kconf.st).layout()
    Tabs(kconf.st).layout()

    grp = getattr(kid, group)

    for id in grp.get_all(False, string, index):
        if fmt:
            print(f"kid.{group}{_formatted_idx(id)} {str(id)}")
        else:
            print(id)

kmviz.add_command(api)
kmviz.add_command(app)
kmviz.add_command(config)
kmviz.add_command(plugin)

if __name__ == "__main__":
    try:
        kmviz()
    except Exception as e:
        kmv_ex(e)



from dash_extensions.enrich import html, Input, Output, State, callback, clientside_callback
from kmviz.core.config import state

from dash import Patch

from kmviz.ui.id_factory import kid
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import khide, kshow
from kmviz.ui.utils import make_select_data

from kmviz.ui.components.option import make_user_option, RangeOption

from dash_iconify import DashIconify

class ConfigLayout:
    def __init__(self, st: state):
        self.st = st
        self.opt = kid.config.new("opt")

    def _db_layout(self):
        layout = []

        for db_name in self.st.engine.list():
            db = self.st.engine.get(db_name)
            children = []

            if db.has_visible_options:
                for opt in db.options.values():
                    if opt.hidden:
                        continue
                    children.append(make_user_option(opt, id=self.opt(f"{db_name}-{opt.name}")))
            layout.append(cf.div(
                kid.config(db_name),
                *children,
                style=kshow if db_name == self.st.defaults.configuration else khide
            ))
        return layout

    def _option_callback(self, name, opt_name):
        def update(value):
            patch = Patch()
            patch[name][opt_name] = value
            return patch
        return update

    def _option_slider_callback(self, opt_name):
        def update(value):
            return [f"{opt_name} = {value}"]
        return update

    def _db_callbacks(self):
        for db_name, db in self.st.engine.all().items():
            for opt_name in db.options:
                if db.options[opt_name].hidden:
                    continue
                callback(
                    Input(self.opt(f"{db_name}-{opt_name}"), "value"),
                    Output(kid.store("options"), "data"),
                    prevent_initial_call=True
                )(self._option_callback(db_name, opt_name))
                if isinstance(db.options[opt_name], RangeOption):
                    idx = self.opt(f"{db_name}-{opt_name}")
                    callback(
                        Input(idx, "value"),
                        Output(f"{idx['type']}-{idx['index']}-{opt_name}", "children"),
                    )(self._option_slider_callback(opt_name))

    def layout(self):
        databases = self.st.engine.list()

        data, value = [], None
        if self.st.defaults.configuration:
            data = make_select_data(self.st.defaults.database)
            value = self.st.defaults.configuration

        show_conf = True

        if self.st.defaults.hide:
            db_name = self.st.defaults.database[0]
            show_conf = self.st.engine.get(db_name).has_visible_options
        else:
            show_conf = any(self.st.engine.get(db_name).has_visible_options for db_name in databases)

        show_select = show_conf and not self.st.defaults.hide

        return cf.div(
            kid.config["div"],
            cf.divider(size="sm", color="gray", label="CONFIGURATION", labelPosition="center"),
            cf.select(
                kid.config["select"],
                clearable=True,
                data=data,
                value=value,
                leftSection=DashIconify(icon="mynaui:config"),
                style=kshow if show_select else khide
            ),
            *self._db_layout(),
            style = kshow if show_conf else khide
        )

    def callbacks(self) -> None:

        clientside_callback(
            """
            function(database, ids) {
                styles = Array(ids.length);
                styles.fill({"display": "none"});
                if (!database) {
                    return styles;
                }
                for (let i = 0; i < ids.length; i++) {
                    if (ids[i]["index"] === database) {
                        styles[i]["display"] = "inline";
                    } else {
                        styles[i]["display"] = "none";
                    }
                }
                return styles;
            }
            """,
            Input(kid.config["select"], "value"),
            State(kid.config.all, "id"),
            Output(kid.config.all, "style"),
            prevent_initial_call=True
        )

        self._db_callbacks()


def make_server_cache_manager(params: dict):
    if params["type"] == "disk":
        from dash_extensions.enrich import FileSystemBackend
        return FileSystemBackend(**params["params"])
    elif params["type"] == "redis":
        from dash_extensions.enrich import RedisBackend
        return RedisBackend(**params["params"])

def make_result_cache_manager(params: dict):
    if params["type"] == "disk":
        from cachelib.file import FileSystemCache
        return FileSystemCache(**params["params"])
    elif params["type"] == "redis":
        from cachelib.redis import RedisCache
        return RedisCache(**params["params"])

def make_callback_cache_manager(params: dict):
    from dash import DiskcacheManager, CeleryManager
    if params["type"] == "disk":
        import diskcache
        return DiskcacheManager(diskcache.Cache(**params["params"]))
    elif params["type"] == "celery":
        from celery import Celery
        return CeleryManager(Celery(**params["params"]))

from kmviz.core import KmVizError

class KmvizResultCache:
    def __init__(self, backend):
        self._backend = backend

    def get(self, uid: str):
        res = self._backend.get(uid)
        if res is None:
            raise KmVizError(f"'{uid}' not in cache")
        return res

    def put(self, uid: str, value):
        self._backend.set(uid, value)

    def inc(self, uid: str, delta):
        return self._backend.inc(uid, delta)

    def dec(self, uid: str, delta):
        return self._backend.dec(uid, delta)

    def has(self, uid: str):
        return self._backend.has(uid)

    def set(self, uid: str, value):
        self._backend.set(uid, value)

    @property
    def back(self):
        return self._backend


from dash import Patch
import dash_mantine_components as dmc
from dash_extensions.enrich import Input, Output, State, html, callback

from kmviz.core.log import kmv_info
from kmviz.core.provider.options import ProviderOption, RangeOption, TextOption, NumericOption, ChoiceOption, MultiChoiceOption

from kmviz.ui.utils import make_select_data

from dash_iconify import DashIconify

def make_range_option(opt: RangeOption, id):
    v = opt.value if opt.value else opt.default
    text_id = id["type"] + "-" + id["index"] + "-" + opt.name
    return html.Div([
        dmc.Text(f"{opt.name} = {v}", size="sm", fw=450, ta="center", id=text_id),
        dmc.Slider(
            id=id,
            min=opt.min,
            max=opt.max,
            step=opt.step,
            value=opt.value if opt.value else opt.default,
            updatemode="mouseup",
            precision=2,
            className="kmviz-dmc-user-slider",
            marks=[
                {"value": opt.min, "label": str(round(opt.min, 2))},
                {"value": opt.max, "label": str(round(opt.max, 2))},
            ],
            size="sm"
        ),
        dmc.Space(h=13)
    ])

def make_numeric_option(opt: NumericOption, id):
    return dmc.NumberInput(
        id=id,
        label=opt.name,
        min=opt.min,
        step=opt.step,
        max=opt.max,
        value=opt.value if opt.value else opt.default,
        precision=2,
        className="kmviz-dmc-user-number-input",
        classNames={"root": "kmviz-dmc-numeric-input-root"},
    )

def make_choice_option(opt: ChoiceOption, id):
    return dmc.Select(
        id=id,
        label=opt.name,
        data=make_select_data(opt.choices),
        value=opt.value if opt.value else opt.default,
        className="kmviz-dmc-user-select",
        clearable=True,
        searchable=True,
        classNames={"root": "kmviz-dmc-select-input-root"},
    )

def make_multichoice_option(opt: MultiChoiceOption, id):
    return dmc.MultiSelect(
        id=id,
        label=opt.name,
        data=make_select_data(opt.choices),
        value=opt.value if opt.value else opt.default,
        className="kmviz-dmc-user-multi-select",
        clearable=True,
        searchable=True,
        classNames={"root": "kmviz-dmc-select-input-root"},
        size="xs"
    )

def make_text_option(opt: TextOption, id):
    return dmc.TextInput(
        id=id,
        label=opt.name,
        value=opt.value if opt.value else opt.default,
        placeholder=opt.placeholder,
        className="kmviz-dmc-user-text-input",
        classNames={"root": "kmviz-dmc-text-input-root"},
        size="xs"
    )

def make_user_option(opt: ProviderOption, id):
    if isinstance(opt, RangeOption):
        return make_range_option(opt, id)
    elif isinstance(opt, NumericOption):
        return make_numeric_option(opt, id)
    elif isinstance(opt, ChoiceOption):
        return make_choice_option(opt, id)
    elif isinstance(opt, MultiChoiceOption):
        return make_multichoice_option(opt, id)
    elif isinstance(opt, TextOption):
        return make_text_option(opt, id)
    return None







from dataclasses import dataclass
from typing import Union, Iterable, Any, Optional

from kmviz.core import KmVizError

class InvalidOptionValue(KmVizError):
    pass

@dataclass
class ProviderOption:
    name: str
    default: Any

    def __post_init__(self):
        if self.value is None:
            self.value = self.default

    @property
    def hidden(self):
        if hasattr(self, "is_hidden"):
            return self.is_hidden
        return False

    @property
    def required(self):
        if hasattr(self, "is_required"):
            return self.is_required
        return False

@dataclass
class NumericOption(ProviderOption):
    min: Optional[Union[int, float]]
    max: Optional[Union[int, float]]
    step: Optional[Union[int, float]]
    value: Union[int, float] = None

@dataclass
class ChoiceOption(ProviderOption):
    choices: Iterable[Union[str, int, float]]
    value: Union[str, int, float] = None

@dataclass
class MultiChoiceOption(ProviderOption):
    choices: Iterable[Union[str, int, float]]
    value: Iterable[Union[str, int, float]] = None

@dataclass
class RangeOption(ProviderOption):
    min: Union[int, float]
    max: Union[int, float]
    step: Union[int, float]
    value: Union[int, float] = None

@dataclass
class TextOption(ProviderOption):
    placeholder: str = None
    value: str = None

def update_option(opt, d):
    for k, v in d.items():
        if hasattr(opt, k):
            setattr(opt, k, v)
        elif (k == "hide"):
            opt.is_hidden = bool(v)
    return opt



from kmviz.core.config import state
from kmviz.ui.id_factory import kid
from kmviz.ui.components.factory import ComponentFactory as cf
from dash_extensions.enrich import callback, Input, State, Output, no_update, clientside_callback
from dash.exceptions import PreventUpdate
from kmviz.ui.utils import prevent_update_on_none
import dash_mantine_components as dmc
import kmviz.core.config as kconf

from dash_iconify import DashIconify

class Index:
    def __init__(self, st: state):
        self.st = st

    def layout(self):
        return cf.div(
            kid.index["div"],
            dmc.Center(cf.h2("Index information")),
            cf.ag_grid(
                kid.index("info-grid"),
                grid_opt = {"domLayout": "autoHeight"}
            ),
            dmc.Space(h=20),
            cf.button(
                kid.index["info-export"],
                "Export",
                disabled=True,
                leftSection=DashIconify(icon="ph:export", width=20)
            ),
            dmc.Center(cf.h2("Index metadata")),
            cf.ag_grid(
                kid.index("meta-grid")
            ),
            dmc.Space(h=20),
            cf.group(
                kid.index["grp-btn"],
                cf.button(
                    kid.index["meta-export"],
                    "Export",
                    disabled=True,
                    leftSection=DashIconify(icon="ph:export", width=20)
                ),
                cf.button(
                    kid.index["meta-rmf"],
                    "Remove filters",
                    disabled=True,
                    leftSection=DashIconify(icon="ph:trash", width=20)
                )
            )
        )

    def callbacks(self) -> None:

        clientside_callback(
            """
            function(n_clicks) {
                if (n_clicks) {
                    return {};
                }
                return window.dash_clientside.no_update;
            }
            """,
            Input(kid.index["meta-rmf"], "n_clicks"),
            Output(kid.index("meta-grid"), "filterModel"),
            prevent_initial_call=True
        )

        clientside_callback(
            """
            function(n_clicks) {
                if (n_clicks) {
                    return true;
                }
                return window.dash_clientside.no_update;
            }
            """,
            Input(kid.index["meta-export"], "n_clicks"),
            Output(kid.index("meta-grid"), "exportDataAsCsv"),
            prevent_initial_call=True
        )

        clientside_callback(
            """
            function(n_clicks) {
                if (n_clicks) {
                    return true;
                }
                return window.dash_clientside.no_update;
            }
            """,
            Input(kid.index["info-export"], "n_clicks"),
            Output(kid.index("info-grid"), "exportDataAsCsv"),
            prevent_initial_call=True
        )

        clientside_callback(
            """
            function(database) {
                if (!database) {
                    return [window.dash_clientside.no_update, window.dash_clientside.no_update, window.dash_clientside.no_update];
                }
                return [false, false, false];
            }
            """,
            Input(kid.kmviz("database"), "value"),
            Output(kid.index["info-export"], "disabled"),
            Output(kid.index["meta-export"], "disabled"),
            Output(kid.index["meta-rmf"], "disabled"),
            prevent_initial_call=True
        )

        @callback(
            Input(kid.kmviz("database"), "value"),
            Output(kid.index("info-grid"), "rowData"),
            Output(kid.index("info-grid"), "columnDefs"),
            Output(kid.index("meta-grid"), "rowData"),
            Output(kid.index("meta-grid"), "columnDefs"),
            prevent_initial_call=True
        )
        def update_index_table(db):
            if db == "__kmviz_df":
                raise PreventUpdate

            p = kconf.st.engine.get(db)

            idf = p.infos_df
            infos_rd = idf.to_dict("records")
            infos_f = [{"field": f} for f in list(idf)]

            df = p.db.df()
            meta_rd = df.to_dict("records")
            meta_f = [{"field": f} for f in list(df)]

            return infos_rd, infos_f, meta_rd, meta_f


from dash import Patch, ALL, MATCH
from dataclasses import dataclass
from typing import Any, List, Dict
from typing_extensions import Self

class IDFactory:
    def __init__(self, name: str) -> None:
        self._name = name
        self._children: List[Self] = []
        self._string_register: Dict[str, Any] = {}
        self._index_register: Dict[str, Any] = {}
        self._register: List[str] = []
        self._index: int=0

    def _make_index_id(self, index: Any) -> Dict[str, Any]:
        if index is None:
            iid = { "type": self._name, "index": self._index }
            self._index += 1
        else:
            idd = { "type": self._name, "index": index}

        r = f"{idd['type']}@{idd['index']}"
        if r not in self._index_register:
            self._index_register[r] = idd
            self._register.append(r)

        return idd

    def _make_string_id(self, index: str) -> str:
        r = f"{self._name}-{index}"
        if r not in self._string_register:
            self._string_register[r] = None
            self._register.append(r)
        return r

    def new(self, suffix: str) -> Self:
        self._children.append(IDFactory(f"{self.name}-{suffix}"))
        return self._children[-1]

    def __getitem__(self, index: str) -> str:
        return self._make_string_id(index)

    def __call__(self, index: str, as_string: bool=False) -> Dict[str, Any]:
        r = self._make_index_id(index)
        if as_string:
            return f"{r['type']}-{r['index']}"
        else:
            return r

    def get_all(self, child: bool=True, s: bool=True, i: bool=True):
        ret = []

        if s:
            for k in self._string_register:
                ret.append(k)
        if i:
            for v in self._index_register.values():
                ret.append(v)

        if child:
            for c in self._children:
                ret.extend(c.get_all(child, s, i))

        return ret

    @property
    def all(self):
        return {"type": self._name, "index": ALL}

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._name

@dataclass
class FactoryStore:
    kmviz: IDFactory = IDFactory("kmviz")
    store: IDFactory = None
    side: IDFactory = None
    tabs: IDFactory = None

    index: IDFactory = None
    table: IDFactory = None
    map: IDFactory = None
    plot: IDFactory = None
    session: IDFactory = None

    input: IDFactory = None
    submit: IDFactory = None
    config: IDFactory = None
    pom: IDFactory = None
    sequence: IDFactory = None
    help: IDFactory = None

    def __post_init__(self):
        self.store = self.kmviz.new("store")
        self.side = self.kmviz.new("side")
        self.tabs = self.kmviz.new("tabs")
        self.plugin = self.kmviz.new("plugin")
        self.index = self.tabs.new("index")
        self.table = self.tabs.new("table")
        self.map = self.tabs.new("map")
        self.plot = self.tabs.new("plot")
        self.sequence = self.tabs.new("sequence")
        self.submit = self.kmviz.new("submit")
        self.input = self.kmviz.new("input")
        self.config = self.kmviz.new("config")
        self.pom = self.side.new("pom")
        self.session = self.kmviz.new("session")
        self.help = self.tabs.new("help")

    def register(self, f: IDFactory, name: str):
        if not hasattr(self, name):
            setattr(self, name, f)

kid = FactoryStore()

def _split_idx(idx):
    if isinstance(idx, dict):
        return idx["type"], idx["index"]
    else:
        return idx.rsplit("-", 1)

def _formatted_idx(idx):
    t, i = _split_idx(idx)

    if isinstance(idx, dict):
        return f'("{i}")'
    else:
        return f'["{i}"]'



import dash_mantine_components as dmc
from dash_extensions.enrich import html, callback, Input, Output, State, Serverside, no_update, clientside_callback
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import khide
from kmviz.ui.components.helpers import load_file_as_df
from kmviz.ui.layouts.input import SideInputLayout, SideDBLayout
from kmviz.ui.layouts.config import ConfigLayout
from kmviz.ui.layouts.submit import SubmitLayout
from kmviz.ui.layouts.notif import NotifLayout

from kmviz.ui.id_factory import kid
from kmviz.ui.utils import icons
from kmviz.core.io import KmVizIOError
from kmviz.core.config import state

import kmviz

class Sidebar:
    def __init__(self, st: state):
        self.st = st
        self.mode = st.mode

        if self.mode == "db":
            self.sinput = SideInputLayout(self.st)
            self.sdb = SideDBLayout(self.st)
            self.sconfig = ConfigLayout(self.st)
            self.ssubmit = SubmitLayout(self.st)
            self.snotif = NotifLayout(self.st)

    def _make_database_layout(self) -> html.Div:
        layout = cf.div(
            kid.kmviz["side-div"],
            dmc.Center(cf.h1(f"kmviz {kmviz.__version_str__}")),
            self.sdb.layout(),
            self.sinput.layout(),
            self.snotif.layout(),
            self.sconfig.layout(),
            self.ssubmit.layout(),
        )

        return layout

    def _make_plot_mode_layout(self) -> html.Div:
        return cf.div(
            kid.pom["div"],
            dmc.Center(cf.h4("Plot Mode")),
            cf.divider(size="sm", color="gray", label="CONFIG", labelPosition="center"),
            cf.group(
                kid.pom["grp-1"],
                cf.text(kid.pom["index"], label="Index", required=True,  classNames={"root": "kmviz-dmc-text-input-root"}),
                cf.text(kid.pom["sep"], label="Separator",  classNames={"root": "kmviz-dmc-text-input-root"})
            ),
            cf.group(
                kid.pom["grp-2"],
                cf.text(kid.pom["latitude"], label="Latitude", classNames={"root": "kmviz-dmc-text-input-root"}),
                cf.text(kid.pom["longitude"], label="Longitude", classNames={"root": "kmviz-dmc-text-input-root"})
            ),
            dmc.Space(h=10),
            dmc.Center(
                cf.upload(
                    kid.pom["upload"],
                    cf.button(kid.pom["load-button"], "Load (csv,tsv,xlsx)", disabled=True, leftSection=icons("file")),
                ),
            ),
            cf.div(kid.pom["error"]),
        )

    def _plot_mode_callbacks(self) -> None:

        @callback(
            Input(kid.pom["index"], "value"),
            Output(kid.pom["load-button"], "disabled"),
            prevent_initial_call=True
        )
        def enable_button(value):
            if value:
                return False
            return True

        @callback(
            Input(kid.pom["upload"], "filename"),
            Input(kid.pom["upload"], "contents"),
            State(kid.pom["index"], "value"),
            State(kid.pom["sep"], "value"),
            State(kid.pom["latitude"], "value"),
            State(kid.pom["longitude"], "value"),
            Output(kid.pom["store"], "data"),
            Output(kid.pom["error"], "children"),
            Output(kid.kmviz["side-layout"], "style"),
            Output(kid.kmviz["main-layout"], "style"),
            prevent_initial_call=True
        )
        def upload_df_file(filename, contents, index, sep, lat, lon):
            try:
                df = load_file_as_df(filename, contents, sep)
                df.rename(columns={index: "ID"}, inplace=True)

                res = dict(df=df, geodata=None)

                if lat and lon:
                    if lat in list(df) and lon in list(df):
                        res["geodata"] = { "latitude": lat, "longitude": lon}
                    else:
                        KmVizIOError(f"'{lat}' or '{lon}' not found in dataframe")

                return Serverside(res), [], khide, {"padding-left": "10px"}

            except KmVizIOError as e:
                no_update, dmc.Text(str(e), color="red", weight=500), no_update, no_update

    def _database_callbacks(self) -> None:
        self.sinput.callbacks()
        self.sdb.callbacks()
        self.sconfig.callbacks()
        self.ssubmit.callbacks()
        self.snotif.callbacks()

    def layout(self) -> html.Div:
        if self.mode == "plot":
            return self._make_plot_mode_layout()
        elif self.mode == "db":
            return self._make_database_layout()

    def callbacks(self) -> None:
        if self.mode == "plot":
            self._plot_mode_callbacks()
        elif self.mode == "db":
            self._database_callbacks()



from dash_extensions.enrich import html, Input, Output, State, callback, clientside_callback
from kmviz.core.config import state

from dash import Patch

from kmviz.ui.id_factory import kid
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import khide, kshow

from kmviz.ui.components.option import make_user_option, RangeOption

from dash_iconify import DashIconify
import dash_mantine_components as dmc

class NotifLayout:
    def __init__(self, st: state):
        self.st = st
        self.opt = kid.config.new("notif")

    def _layout(self):
        layout = []
        for name, opt in self.st.notif.options().items():
            layout.append(cf.div(
                self.opt(name),
                make_user_option(opt, id=self.opt(f"notif-{opt.name}"))
            ))
        return layout

    def _option_callback(self, opt_name):
        def update(value):
            patch = Patch()
            patch["notif"][opt_name] = value
            return patch
        return update

    def _callbacks(self):
        for opt_name in self.st.notif.options().keys():
            callback(
                Input(self.opt(f"notif-{opt_name}"), "value"),
                Output(kid.store("options"), "data"),
                prevent_initial_call=True
            )(self._option_callback(opt_name))


    def layout(self):
        if not len(self.st.notif.options()):
            return cf.div(kid.config["notif-div"], style = {"display": "none"})

        return cf.div(
            kid.config["div"],
            cf.divider(size="sm", color="gray", label="NOTIFICATION", labelPosition="center"),
            *self._layout(),
            dmc.Space(h=10),
        )

    def callbacks(self) -> None:
        self._callbacks()


from .db import MetaDB
import duckdb
from typing import List

class ParquetDB(MetaDB):
    def __init__(self, idx: str, geodata, files: List[str]):
        super().__init__(idx, geodata)
        self._files = files

    def connect(self) -> None:
        return None

    def query(self, keys):
        keys_str = (f"'{x}'" for x in keys)
        db = duckdb.read_parquet(self._files)
        Id = self.idx
        Q = f"""
            SELECT *
            FROM db
            WHERE {Id} IN ({','.join(keys_str)})
        """
        d = duckdb.sql(Q).df()
        d.rename(columns={self.idx: "ID"}, inplace=True)
        return d

    def df(self):
        return None

    def keys(self):
        return None


from .tsv import TsvMetaDB
from .sqlite import SQLiteMetaDB
from .mysql import MySQLMetaDB
from .parquet import ParquetDB
from .db import MetaDB
from kmviz.core import KmVizError

METADBS = {
    "tsv": TsvMetaDB,
    "sqlite": SQLiteMetaDB,
    "mysql": MySQLMetaDB,
    "parquet": ParquetDB
}




